"""
Create downloads registered for a clld app.
"""
from clld.interfaces import IDownload
from clld.cliutil import BootstrappedAppConfig


def register(parser):
    parser.add_argument(
        "config_uri", action=BootstrappedAppConfig, help="ini file providing app config")
    parser.add_argument(
        'domain',
        help="domain name under which the app is served. Necessary to create correct URLs "
             "in RDF downloads.",
    )
    parser.add_argument(
        '-l', '--list', default=False, action='store_true', help='only list registered downloads')


def run(args):
    """
    Create all registered downloads (locally).
    """
    if args.domain:
        args.env['request'].environ['HTTP_HOST'] = args.domain
    for name, download in args.env['registry'].getUtilitiesFor(IDownload):  # pragma: no cover
        args.log.info('creating download %s' % name)
        if not args.list:
            download.create(args.env['request'])



import collections

from clld.db.models import common
from clld import interfaces

__version__ = "11.1.0"


class Resource(collections.namedtuple('Resource', 'name model interface with_index with_rdfdump')):

    def __new__(cls, name, model, interface, with_index=True, with_rdfdump=True):
        return super(Resource, cls).__new__(cls, name, model, interface, with_index, with_rdfdump)

    @property
    def plural(self):
        return self.name + 's'


RESOURCES = [
    Resource('dataset', common.Dataset, interfaces.IDataset, with_index=False),
    Resource('contribution', common.Contribution, interfaces.IContribution),
    Resource('parameter', common.Parameter, interfaces.IParameter),
    Resource('language', common.Language, interfaces.ILanguage),
    Resource('contributor', common.Contributor, interfaces.IContributor),
    Resource('source', common.Source, interfaces.ISource),
    Resource('sentence', common.Sentence, interfaces.ISentence),
    Resource('valueset', common.ValueSet, interfaces.IValueSet),
    Resource('value', common.Value, interfaces.IValue),
    Resource('unitparameter', common.UnitParameter, interfaces.IUnitParameter),
    Resource('unit', common.Unit, interfaces.IUnit),
    Resource('unitvalue', common.UnitValue, interfaces.IUnitValue),
    Resource(
        'combination',
        common.Combination,
        interfaces.ICombination,
        with_index=False,
        with_rdfdump=False),
]




def test_add_renderer_globals(env, mocker):
    from clld.web.subscribers import add_renderer_globals

    ctx = {'renderer_name': 'path/base.ext.mako', 'request': None}
    add_renderer_globals(mocker.Mock(path_base_ext=lambda **kw: {'a': 3}), ctx)
    assert ctx['a'] == 3



from clld.web.util.htmllib import HTML

__all__ = ['BASE_URL', 'get_doi', 'url', 'logo', 'link', 'badge']

BASE_URL = 'https://doi.org'


def get_doi(obj):
    if isinstance(obj, str):
        return obj
    return getattr(obj, 'doi', getattr(obj, 'jsondata', {}).get('doi'))


def url(obj):
    doi = get_doi(obj)
    if doi:
        return '{0}/{1}'.format(BASE_URL, doi)


def logo(req, width='20'):
    return HTML.img(
        src=req.static_url('clld:web/static/images/doi_logo.png'),
        width=width,
        style="margin-top: -2px")


def link(req, obj, label=None, with_logo=True):
    doi = get_doi(obj)
    akw = dict(title='DOI: {}'.format(doi), href=url(doi))
    return HTML.a(logo(req) if with_logo else '', label or doi, **akw)


def badge(obj):
    doi = get_doi(obj)
    if doi.startswith('10.5281/zenodo.'):
        return HTML.a(
            HTML.img(
                src="https://zenodo.org/badge/DOI/{}.svg".format(doi), alt='DOI: {}'.format(doi)),
            href=url(doi))
    return HTML.a(doi, href=url(doi))



import pytest

from clld.lib.bibtex import *


@pytest.mark.parametrize(
    "arg,res",
    [
        (bytes("\\ss \xef".encode('latin1')), 'ß\xef'),
        ("\\ss ", 'ß'),
        ('\u2013', '\u2013'),
        ('?[\\u65533]', '\ufffd'),
    ])
def test_unescape(arg, res):
    assert unescape(arg) == res


def test_u_unescape():
    assert u_unescape('?[\\u123] ?[\\u1234]') == '{ \u04d2'


def test_stripctrlchars():
    assert stripctrlchars('a\u0008\u000ba') == 'aa'
    assert stripctrlchars(None) is None


def test_Record_author_editor():
    rec = Record('article', '1', author=['a', 'b'], editor='a and b')
    assert rec['author'] == 'a and b'
    assert rec['editor'] == rec.get('editor')
    assert rec.getall('editor') == ['a', 'b']


def test_Record(mocker):
    rec = Record('article', '1', author=['a', 'b'], editor='a and b')
    rec.format('txt')
    Record.from_string(str(rec), lowercase=True)
    Record.from_object(mocker.Mock())


@pytest.mark.parametrize(
    'genre,fields,expected_genre,substring',
    [
        (
            'incollection',
            dict(
                title='The Title',
                editor='ed',
                booktitle='bt',
                school='s',
                issue='i',
                pages='1-4',
                publisher='M',
                note="Revised edition"),
            EntryType.incollection,
            'In '),
        (
            'article',
            dict(
                title='The Title',
                journal='The Journal',
                volume="The volume",
                issue='issue'),
            EntryType.article,
            'The Journal'),
        ('xmisc', dict(note='Something'), EntryType.misc, 'Something'),
        (
            'book',
            dict(
                title='The Title',
                author='author',
                editor='ed',
                booktitle='bt',
                school='s',
                issue='i',
                pages='1-4',
                publisher='M',
                note="Revised edition"),
            EntryType.book,
            'bt'),
    ]
)
def test_Record_text(genre,fields,expected_genre,substring):
    rec = Record(genre, '1', **fields)
    assert expected_genre == rec.genre
    assert substring in rec.text()


def test_Database(testsdir):
    db = Database([])
    assert not len(db)
    db = Database([Record('book', 'id')])
    assert db[0] == db['id']
    assert str(db)
    db = Database.from_file('notexisting.bib')
    assert not len(db)
    db = Database.from_file(testsdir / 'test.bib')
    assert len(db) == 1
    assert '@' in db[0]['title']
    assert [r for r in db]
    with pytest.raises(NotImplementedError):
        db.format('txt')



"""Shared functionality for clld console scripts."""
import pathlib
import argparse
import warnings
import functools
import importlib
import collections

from sqlalchemy import engine_from_config
from pyramid.paster import get_appsettings, bootstrap
from nameparser import HumanName
from clldutils.misc import slug
from clldutils import clilib

from clld.db.meta import DBSession, Base
from clld.db.models import common
from clld.lib import bibtex

__all__ = [
    'AppConfig',
    'BootstrappedAppConfig',
    'SessionContext',
    'data_file',
    'add_language_codes',
    'bibtex2source',
    'confirm',
    'Data']

confirm = functools.partial(clilib.confirm, default=False)


# Moved here from distutils.util, due to this package being deprecated.
def strtobool(val):  # pragma: no cover
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return 1
    if val in ('n', 'no', 'f', 'false', 'off', '0'):
        return 0
    raise ValueError("invalid truth value %r" % (val,))


class AppConfig(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        p = pathlib.Path(values.split('#')[0])
        if not (p.exists() and p.is_file()):
            raise ValueError()
        setattr(namespace, self.dest, p)
        namespace.module = p.resolve().parent.name
        namespace.settings = get_appsettings(values)

        if namespace.module == 'tests':
            namespace.module = 'clld'

        namespace.module = importlib.import_module(namespace.module)
        namespace.data_file = functools.partial(data_file, namespace.module)
        namespace.module_dir = pathlib.Path(namespace.module.__file__).parent
        namespace.migrations_dir = namespace.module_dir.joinpath('..', 'migrations')


class BootstrappedAppConfig(AppConfig):
    def __call__(self, parser, namespace, values, option_string=None):
        AppConfig.__call__(self, parser, namespace, values, option_string=option_string)
        namespace.env = bootstrap(values)
        try:
            namespace.initializedb = importlib.import_module(
                '.'.join([namespace.module.__name__, 'scripts', 'initializedb']))
        except ImportError as e:
            warnings.warn(str(e))
            namespace.initializedb = None


class SessionContext:
    """
    To support accessing configured sessions in cli commands.
    """
    def __init__(self, settings):
        self.settings = {'sqlalchemy.url': settings} if isinstance(settings, str) else settings
        self.engine = None

    def __enter__(self):
        self.engine = engine_from_config(self.settings)
        DBSession.remove()
        DBSession.configure(bind=self.engine)
        assert DBSession.bind == self.engine
        Base.metadata.create_all(self.engine)
        return DBSession

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.engine:
            self.engine.dispose()
        DBSession.remove()


def add_language_codes(data, lang, isocode, glottocodes=None, glottocode=None):
    def identifier(type_, id_):
        return data.add(
            common.Identifier, '%s:%s' % (type_, id_),
            id='%s:%s' % (type_, id_),
            name=id_,
            type=getattr(common.IdentifierType, type_).value)

    if isocode and len(isocode) == 3:
        DBSession.add(common.LanguageIdentifier(
            language=lang, identifier=identifier('iso', isocode)))

    if glottocode or (glottocodes and isocode and isocode in glottocodes):
        glottocode = glottocode or glottocodes[isocode]
        DBSession.add(common.LanguageIdentifier(
            language=lang, identifier=identifier('glottolog', glottocode)))


def bibtex2source(rec, cls=common.Source, lowercase_id=False):
    year, fields, jsondata = bibtex.unescape(rec.get('year', 'nd')), {}, {}
    for field in bibtex.FIELDS:
        if field in rec:
            value = bibtex.unescape(rec[field])
            container = fields if hasattr(cls, field) else jsondata
            container[field] = value

    etal, eds = '', ''
    authors = rec.get('author')
    if not authors:
        authors = rec.get('editor', '')
        if authors:
            eds = ' (eds.)'
    if authors:
        authors = bibtex.unescape(authors).split(' and ')
        if len(authors) > 2:
            authors = authors[:1]
            etal = ' et al.'

        authors = [n.last or n.first for n in [HumanName(a) for a in authors]]
        authors = '%s%s%s' % (' and '.join(authors), etal, eds)

    return cls(
        id=slug(rec.id, lowercase=lowercase_id),
        name=('%s %s' % (authors, year)).strip(),
        description=bibtex.unescape(rec.get('title', rec.get('booktitle', ''))),
        jsondata=jsondata,
        bibtex_type=rec.genre,
        **fields)


def data_file(module, *comps):
    """Return Path object of file in the data directory of an app."""
    return pathlib.Path(module.__file__).parent.joinpath('..', 'data', *comps)


class Data(collections.defaultdict):

    """Dictionary, serving to store references to new db objects during data imports.

    The values are dictionaries, keyed by the name of the model class used to create the
    new objects.

    >>> data = Data()
    >>> l = data.add(common.Language, 'l', id='abc', name='Abc Language')
    >>> assert l == data['Language']['l']
    """

    def __init__(self, **kw):
        super(Data, self).__init__(dict)
        self.defaults = kw

    def add(self, model_, key_, **kw):
        """
        Create an instance of a model class to be persisted in the database.

        :param model_: The model class we want to create an instance of.
        :param key_: A key which can be used to retrieve the instance later.
        :param kw: Keyword parameters passed to model class for initialisation.
        :return: The newly created instance of model class.
        """
        if '.' in kw.get('id', ''):
            raise ValueError('Object id contains illegal character "."')
        if list(kw.keys()) == ['_obj']:
            # if a single keyword parameter _obj is passed, we take it to be the object
            # which should be added to the session.
            new = kw['_obj']
        else:
            for k, v in self.defaults.items():
                kw.setdefault(k, v)
            new = model_(**kw)
        self[model_.__name__][key_] = new
        DBSession.add(new)
        return new



"""Default DataTable for Source objects."""
from clld.db.models.common import Language, LanguageSource, Source
from clld.web.datatables.base import DataTable, Col, LinkCol, DetailsRowLinkCol
from clld.lib.bibtex import EntryType
from clld.web.util.htmllib import HTML


class TypeCol(Col):

    """Render the BibTeX type of a Source item."""

    def __init__(self, dt, name, *args, **kw):
        kw['sTitle'] = 'BibTeX type'
        kw['choices'] = [(t.value, t.description) for t in EntryType]
        super(TypeCol, self).__init__(dt, name, *args, **kw)

    def format(self, item):
        return getattr(item.bibtex_type, 'description', '')

    def order(self):
        return Source.bibtex_type

    def search(self, qs):
        return Source.bibtex_type == getattr(EntryType, qs)


class Sources(DataTable):

    """Default DataTable for Source objects."""

    __constraints__ = [Language]
    __toolbar_kw__ = {'dl_formats': {'bib': 'BibTeX'}}

    def base_query(self, query):
        if self.language:
            query = query.join(LanguageSource)\
                .filter(LanguageSource.language_pk == self.language.pk)
        return query

    def col_defs(self):
        return [
            DetailsRowLinkCol(self, 'd'),
            LinkCol(self, 'name'),
            Col(self, 'description', sTitle='Title', format=lambda i: HTML.span(i.description)),
            Col(self, 'year'),
            Col(self, 'author'),
            TypeCol(self, 'bibtex_type'),
        ]



import pytest
from sqlalchemy.sql.expression import cast
from sqlalchemy.types import Integer
from zope.interface import Interface, classImplements

from clld.db.models import common
from clld.web.datatables.base import (
    DataTable, Col, LinkCol, DetailsRowLinkCol, LinkToMapCol, IntegerIdCol, IdCol,
    ExternalLinkCol, filter_number, PercentCol,
)

from testutils import handle_dt


class Table(DataTable):
    __toolbar_kw__ = dict(exclude=['atom'])

    def col_defs(self):
        return [
            Col(self, 'pk'),
            DetailsRowLinkCol(self, 'd'),
            LinkToMapCol(self, 'm'),
            LinkCol(self, 'link'),
            Col(self, 'active'),
            Col(self, 'name', model_col=None, sDescription='MyName'),
            Col(self, 'description', format=lambda i: 'x'),
            IntegerIdCol(self, 'id'),
            IdCol(self, 'nid',
                  get_object=lambda i: i, model_col=common.Contributor.id),
            ExternalLinkCol(self, 'url'),
            Col(self, 'id', choices=['x', 'y'], select='multiple', sTooltip='xytt'),
        ]


def test_DataTable(env, request_factory):
    dt = Table(env['request'], common.Contributor)
    assert 'exclude' in dt._toolbar.options
    assert str(dt) == 'Contributors'
    assert repr(dt) == 'Contributors'
    dt.get_query(undefer_cols=['updated']).all()
    html = dt.render()
    assert 'xytt' in html
    assert ' multiple ' in html
    assert 'MyName' in html


@pytest.mark.parametrize('params', [
    {},
    {
        'sSearch_0': '1..1000000',
        'sSearch_4': 'True',
        'sSearch_6': 'e',
        'iSortingCols': '1',
        'iSortCol_0': '0',
        'sSortDir_0': 'desc',
    },
    {'iSortingCols': 'x'},
    {'iSortingCols': '1', 'iSortCol_0': '7'},
    {'sSearch_7': '1'},
    {'sSearch_8': 'id'},
    {'iDisplayLength': 'nonnumber'},
    {'iDisplayStart': 'nonnumber'},
])
def test_DataTable_cols(request_factory, params):
    with request_factory(params=params) as req:
        handle_dt(req, Table, common.Contributor)


def test_DataTable2(request_factory):
    class TestCol(Col):
        def search(self, qs):
            return filter_number(cast(self.dt.model.pk, Integer), qs, type_=int)

    class TestTable(DataTable):
        def col_defs(self):
            return [Col(self, 'latitude'), TestCol(self, 'pk')]

    with request_factory(params={'sSearch_0': '1..2,3..4'}) as req:
        handle_dt(req, TestTable, common.Language)

    with request_factory(params={'sSearch_0': '> 1', 'sSearch_1': '> 1'}) as req:
        handle_dt(req, TestTable, common.Language)


def test_DataTable3(env, request_factory):
    class TestTable(DataTable):
        def col_defs(self):
            return [PercentCol(self, 'frequency'), Col(self, 'name', choices=['a'])]

    handle_dt(env['request'], TestTable, common.Value)

    with request_factory(params={'sSearch_0': '> 10', 'sSearch_1': 'a'}) as req:
        handle_dt(req, TestTable, common.Value)


def test_DataTable_for_custom_class(env):
    from clld.web.datatables.base import DataTable

    class A(object):  # no route is registered for the name 'as' ...
        pass

    class ILanguage(Interface):  # but there is one for 'languages'!
        pass

    classImplements(A, ILanguage)
    dt = DataTable(env['request'], A)
    assert 'languages' in dt.options['sAjaxSource']



import pytest

from clld.lib.bibtex import Record
from clld.lib.coins import ContextObject


@pytest.mark.parametrize(
    'sid,mtx,data,assertion',
    [
        ('sid', 'journal', [('jtitle', '\xe2')], lambda sa: '%C3%A2' in sa['title']),
        (
            'sid',
            'journal',
            [('jtitle', bytes('ä'.encode('utf8')))],
            lambda sa: '%C3%A4' in sa['title']),
        ('sid', 'book', [('btitle', 'the title')], lambda sa: bool(sa)),
        ('äöü', 'äöü', [], lambda sa: bool(sa)),
        ('äöü'.encode('latin1'), None, [], lambda sa: bool(sa)),
    ]
)
def test_ContextObject_span_attrs(sid,mtx,data,assertion):
    assert assertion(ContextObject(sid, mtx, *data).span_attrs())


@pytest.mark.parametrize(
    'genre,fields',
    [
        ('book', dict(title='The Title', author='L, F')),
        ('article', dict(title='The Title', author='The One and The Other', journal='J')),
        ('phdthesis', dict(title='The Title')),
        ('conference', dict(title='The Title', booktitle='something')),
    ]
)
def test_ContextObject_from_bibtex(genre, fields):
    co = ContextObject.from_bibtex('sid', Record(genre, '1', **fields))
    assert isinstance(co.span_attrs(), dict)



import pytest
from pyramid.paster import get_appsettings

import clld
from clld.lib.bibtex import Record
from clld.cliutil import *
from clld.db.models.common import Language


def test_data_file():
    assert data_file(clld, 'util.py').stem == 'util'


def test_SessionContext(testsdir):
    settings = get_appsettings('{0}#main'.format(testsdir / 'test.ini'))
    with SessionContext(settings):
        pass


def test_bibtex2source():
    bibtex2source(Record('book', 'id', author='M, R and G, H and Z, U'))
    bibtex2source(Record('book', 'id', editor='M, R and G, H'))
    bibtex2source(Record('book', 'id', title='tb', customfield='cf', year="1920}"))
    assert bibtex2source(Record('misc', 'Id', title='title')).id == 'Id'


def test_Data(mocker):
    session = set()
    mocker.patch('clld.cliutil.DBSession', session)
    d = Data(jsondata={})
    d.add(Language, 'l', id='l', name='l')
    assert session
    d.add(Language, 'l2', _obj=5)
    with pytest.raises(ValueError):
        d.add(Language, 'l3', id='l.3')


def test_add_language_codes(env):
    add_language_codes(Data(), Language(), 'iso', glottocodes=dict(iso='glot1234'))



"""Database utilities."""
import re
import time
import functools

from sqlalchemy import Integer
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import cast
import transaction

from clld.db.meta import DBSession
from clld.db.models import common

__all__ = [
    'as_int', 'contains', 'icontains', 'compute_language_sources', 'compute_number_of_values',
    'get_distinct_values', 'page_query']


def as_int(col):
    return cast(col, Integer)


def _contains(method, col, qs):
    """Infix search condition.

    Basic support is provided for specifying matches at beginning or end of the text.
    In addition, every special search syntax recognized by the ``LIKE`` operator of the
    underlying SQL dialect will work. Thus, to match strings with a specific <prefix>
    and <suffix>, a query string of the form "^<prefix>%<suffix>$" will do on PostgreSQL.

    .. seealso:: https://www.postgresql.org/docs/9.1/static/functions-matching.html
    """
    spattern = re.compile(r'^(\^|\\b)')
    epattern = re.compile(r'(\$|\\b)$')

    prefix, suffix = '%', '%'
    if spattern.search(qs):
        qs, prefix = spattern.sub('', qs), ''

    if epattern.search(qs):
        qs, suffix = epattern.sub('', qs), ''

    # Prevent invalid LIKE patterns:
    if qs.endswith('\\') and not qs.endswith('\\\\'):
        qs += '\\'
    return getattr(col, method)(prefix + qs + suffix)


icontains = functools.partial(_contains, 'ilike')
contains = functools.partial(_contains, 'like')


def compute_language_sources(*references):
    """compute relations between languages and sources.

    by going through the relevant models derived from the HasSource mixin.
    """
    old_sl = {(ls.source_pk, ls.language_pk) for ls in DBSession.query(common.LanguageSource)}

    references = list(references)
    references.extend([
        (common.ValueSetReference, 'valueset'),
        (common.SentenceReference, 'sentence')])

    for model, attr in references:
        for ref in DBSession.query(model):
            source_pk = ref.source_pk
            language_pk = getattr(ref, attr).language_pk
            if (source_pk, language_pk) not in old_sl:
                old_sl.add((source_pk, language_pk))
                DBSession.add(common.LanguageSource(
                    language_pk=language_pk, source_pk=source_pk))


def compute_number_of_values():
    """compute number of values per valueset and store it in valueset's jsondata."""
    for valueset in DBSession.query(common.ValueSet).options(
        joinedload(common.ValueSet.values)
    ):
        valueset.update_jsondata(_number_of_values=len(valueset.values))


def get_distinct_values(col, key=None):
    return sorted((c for c, in DBSession.query(col).distinct() if c), key=key)


def page_query(q, n=1000, verbose=False, commit=False):
    """Go through query results in batches.

    .. seealso:: http://stackoverflow.com/a/1217947
    """
    s = time.time()
    offset = 0
    while True:
        r = False
        for elem in q.limit(n).offset(offset):
            r = True
            yield elem
        if commit:  # pragma: no cover
            transaction.commit()
            transaction.begin()
        offset += n
        e = time.time()
        if verbose:
            print(e - s, offset, 'done')  # pragma: no cover
        s = e
        if not r:
            break



import re
import json
import io

import pytest

from pyramid.renderers import render
from rdflib import Graph, URIRef
import html5lib

from clld import RESOURCES
from clld.lib import rdf
from clld.db.models.common import Parameter


_RESOURCES = [_rsc for _rsc in RESOURCES if _rsc.name != 'testresource']


@pytest.mark.parametrize(
    'rsc',
    [r for r in _RESOURCES if hasattr(r.model, 'first')],
)
def test_detail_html(request_factory, rsc):
    with request_factory(matched_route='home', map=None) as req:
        res = render(
            '%s/detail_html.mako' % rsc.name, {'ctx': rsc.model.first()}, request=req)
        html5lib.parse(res)
        if rsc.name == 'dataset':
            assert 'http://example.org/privacy' in res
            assert 'Privacy Policy' in res


@pytest.mark.parametrize(
    'rsc',
    [r for r in _RESOURCES if hasattr(r.model, 'first') and r.name != 'dataset'],
)
def test_index_html(request_factory, rsc):
    with request_factory(matched_route='home', map=None) as req:
        dt = req.get_datatable(rsc.name + 's', rsc.model)
        res = render('%s/index_html.mako' % rsc.name, {'ctx': dt}, request=req)
        html5lib.parse(res)


@pytest.mark.parametrize(
    'rsc',
    [r for r in _RESOURCES if hasattr(r.model, 'first')],
)
def test_json(env, rsc):
    res = render('json', dict(ctx=rsc.model.first()), request=env['request'])
    json.loads(res)


def test_rdf(env):
    global RESOURCES
    RESOURCES = _RESOURCES
    qname_as_resource = re.compile(r'rdf:[a-z]+=\"\w+:\w+\"')
    for rsc in _RESOURCES:
        if not hasattr(rsc.model, 'first'):
            continue
        ctx = rsc.model.first()
        res = render('%s/rdf.mako' % rsc.name, dict(ctx=ctx), request=env['request'])
        assert not qname_as_resource.search(res)
        g = Graph()
        g.parse(io.StringIO(res), format='xml')
        for predicate in ['void:inDataset', 'skos:prefLabel']:
            if predicate == 'void:inDataset' and rsc.name == 'dataset':
                continue
            subject = URIRef(env['request'].resource_url(ctx))
            predicate = URIRef(rdf.url_for_qname(predicate))
            assert (subject, predicate, None) in g
    p = Parameter.get('parameter')
    res = render('parameter/rdf.mako', dict(ctx=p), request=env['request'])
    assert p.domain[0].name in res



"""We provide some infrastructure to build extensible database models."""
import json
import sqlite3

from sqlalchemy import Column, Integer, String, Boolean, DateTime, func, event
from sqlalchemy.pool import Pool
from sqlalchemy.orm import (
    declarative_base, declared_attr, scoped_session, sessionmaker, deferred, undefer,
)
from sqlalchemy.exc import NoResultFound, MultipleResultsFound, DisconnectionError
from sqlalchemy.inspection import inspect
from sqlalchemy.types import TypeDecorator, VARCHAR

import zope.sqlalchemy
from clldutils.misc import NO_DEFAULT
from clldutils import jsonlib


@event.listens_for(Pool, "checkout")
def ping_connection(dbapi_connection, connection_record, connection_proxy):
    """Event listener to handle disconnects.

    Implements
    `pessimistic disconnect handling <http://docs.sqlalchemy.org/en/rel_0_9/core/\
    pooling.html#disconnect-handling-pessimistic>`_.

    .. note::

        Our implementation is mildly dialect specific, but works for sqlite and
        PostgreSQL. For oracle, the 'ping' query should read *SELECT 1 FROM DUAL* or
        similar.
    """
    cursor = dbapi_connection.cursor()
    try:
        cursor.execute("SELECT 1")
        if not isinstance(dbapi_connection, sqlite3.Connection):  # pragma: no cover
            cursor.execute("SET default_text_search_config TO 'english'")
    except:  # noqa: E722; # pragma: no cover
        # dispose the whole pool instead of invalidating one at a time
        connection_proxy._pool.dispose()

        # raise DisconnectionError - pool will try
        # connecting again up to three times before raising.
        raise DisconnectionError()
    cursor.close()


DBSession = scoped_session(sessionmaker())
zope.sqlalchemy.register(DBSession)


class JSONEncodedDict(TypeDecorator):

    """Represents an immutable structure as a json-encoded string.
    Loads/serializes an empty dict for any empty value.
    """

    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        return json.dumps(value or {})

    def process_result_value(self, value, dialect):
        return json.loads(value) if value else {}


class Base(declarative_base()):

    """The declarative base for all our models."""

    __abstract__ = True

    @declared_attr
    def __tablename__(cls):
        """We derive the table name from the model class name.

        This should be safe,
        because we don't want to have model classes with the same name either.
        Care has to be taken, though, to prevent collisions with the names of tables
        which are automatically created (history tables for example).
        """
        return cls.__name__.lower()

    #: All our models have an integer primary key which has nothing to do with
    #: the kind of data stored in a table. 'Natural' candidates for primary keys
    #: should be marked with unique constraints instead. This adds flexibility
    #: when it comes to database changes.
    pk = Column(Integer, primary_key=True, doc='primary key')

    #: To allow for timestamp-based versioning we store a timestamp for creation or an object.
    @declared_attr
    def created(cls):
        return deferred(Column(DateTime(timezone=True), default=func.now()))

    #: Timestamp for latest update of an object.
    @declared_attr
    def updated(cls):
        return deferred(Column(DateTime(timezone=True), default=func.now(), onupdate=func.now()))

    #: The active flag is meant as an easy way to mark records as obsolete or inactive,
    #: without actually deleting them.
    @declared_attr
    def active(cls):
        return deferred(Column(Boolean, default=True))

    #: To allow storage of arbitrary key,value pairs with typed values, each model
    #: provides a column to store JSON encoded dicts.
    jsondata = Column(JSONEncodedDict)

    def __init__(self, jsondata=None, **kwargs):
        kwargs['jsondata'] = jsondata or {}
        super(Base, self).__init__(**kwargs)

    def update_jsondata(self, **kw):
        """Convenience function.

        Since we use the simple
        `JSON encoded dict recipe <http://docs.sqlalchemy.org/en/rel_0_9/core/types.html\
        #marshal-json-strings>`_
        without mutation tracking, we provide a convenience method to update
        """
        d = (self.jsondata or {}).copy()
        d.update(kw)
        self.jsondata = d

    @property
    def jsondatadict(self):
        """Deprecated convenience function.

        Use jsondata directly instead, which is guaranteed to be a dictionary.
        """
        return self.jsondata or {}

    @property
    def replacement_id(self):
        """Used to allow automatically redirecting to a 'better' version of a resource."""
        if not self.active:
            return self.jsondata.get('__replacement_id__')

    @classmethod
    def get(cls, value, key=None, default=NO_DEFAULT, session=None):
        """Convenience method to query a model where exactly one result is expected.

        e.g. to retrieve an instance by primary key or id.

        :param value: The value used in the filter expression of the query.
        :param str key: The key or attribute name to be used in the filter expression. If\
        None is passed, defaults to *pk* if value is ``int`` otherwise to *id*.
        """
        session = session or DBSession
        if key is None:
            key = 'pk' if isinstance(value, int) else 'id'
        try:
            return session.query(cls)\
                .options(undefer('updated')).filter_by(**{key: value}).one()
        except (NoResultFound, MultipleResultsFound):
            if default is NO_DEFAULT:
                raise
            return default

    @classmethod
    def first(cls):
        """More convenience."""
        return DBSession.query(cls).order_by(cls.pk).first()

    def __json__(self, req):
        """Custom JSON serialization of an object.

        :param req: pyramid Request object.
        :return: ``dict`` suitable for serialization as JSON.
        """
        exclude = {'active', 'version', 'created', 'updated', 'polymorphic_type'}
        cols = [
            col.key for om in inspect(self).mapper.iterate_to_root()
            for col in om.local_table.c
            if col.key not in exclude and not exclude.add(col.key)]
        return {col: jsonlib.format(getattr(self, col)) for col in cols}

    def __str__(self):
        """A human readable label for the object."""
        r = getattr(self, 'name', None)
        if not r:
            r = getattr(self, 'id', None)
        if not r:
            r = '%s%s' % (self.__class__.__name__, self.pk)
        return r

    def __repr__(self):
        return '<%s %r>' % (
            self.__class__.__name__, getattr(self, 'id', self.pk))


class PolymorphicBaseMixin(object):

    """Mixin providing the wiring for joined table inheritance.

    We use joined table inheritance to allow projects to augment base ``clld``
    models with project specific attributes. This mixin class prepares
    models to serve as base classes for inheritance.
    """

    polymorphic_type = Column(String(20))

    @declared_attr
    def __mapper_args__(cls):
        return {
            'polymorphic_on': cls.polymorphic_type,
            'polymorphic_identity': 'base',
            'with_polymorphic': '*',
        }


def is_base(cls):
    """Determine whether a class is a base class or an inheriting one.

    :param cls: Model class.
    :return: ``bool`` signaling whether ``cls`` is a base class or derived, i.e.\
    customized.
    """
    # replace with inspection?
    # see http://docs.sqlalchemy.org/en/rel_0_9/orm/mapper_config.html
    # ?highlight=polymorphic_identity#sqlalchemy.orm.mapper.Mapper.polymorphic_identity
    return PolymorphicBaseMixin in cls.__bases__


class CustomModelMixin(object):

    """Mixin for customized classes in our joined table inheritance scheme.

    .. note::

        With this scheme there can be only one specialized mapper class per inheritable
        base class.
    """

    @declared_attr
    def __mapper_args__(cls):
        return {'polymorphic_identity': 'custom'}  # pragma: no cover



"""Provide webhelpers2.html functionality in a backwards compatible way.

We reverted back to using a third-party library. This module does only make sure, import
statements still work as before.
"""
from clldutils.html import literal, HTML, escape
assert literal and HTML and escape



"""Default DataTable for Unit objects."""
from sqlalchemy.orm import joinedload

from clld.db.models.common import Unit, Language
from clld.web.datatables.base import DataTable, LinkCol


class DescriptionLinkCol(LinkCol):

    """Render a link to the unit using the description as label."""

    def get_attrs(self, item):
        return {'label': item.description}


class Units(DataTable):

    """Default DataTable for Unit objects."""

    __constraints__ = [Language]

    def base_query(self, query):
        query = query.join(Language).options(joinedload(Unit.language))

        if self.language:
            return query.filter(Unit.language == self.language)
        return query

    def col_defs(self):
        return [
            LinkCol(self, 'name'),
            DescriptionLinkCol(self, 'description'),
            LinkCol(
                self, 'language', model_col=Language.name, get_obj=lambda i: i.language),
        ]



from sqlalchemy import Column, Integer, Unicode, ForeignKey, UniqueConstraint
from sqlalchemy.orm import relationship, backref, joinedload
from sqlalchemy.ext.declarative import declared_attr

from zope.interface import implementer

from clld.db.meta import Base, PolymorphicBaseMixin
from clld import interfaces

from . import (
    IdNameDescriptionMixin,
    DataMixin, HasDataMixin, FilesMixin, HasFilesMixin,
    HasSourceNotNullMixin,
    parameter as _parameter)  # needed to initalize relationship
assert _parameter

__all__ = ('ValueSet', 'ValueSetReference')


class ValueSet_data(Base, DataMixin):
    pass


class ValueSet_files(Base, FilesMixin):
    pass


@implementer(interfaces.IValueSet)
class ValueSet(Base,
               PolymorphicBaseMixin,
               IdNameDescriptionMixin,
               HasDataMixin,
               HasFilesMixin):

    """The intersection of Language, Parameter, and optionally Contribution."""

    __table_args__ = (
        UniqueConstraint('language_pk', 'parameter_pk', 'contribution_pk'),
    )

    language_pk = Column(Integer, ForeignKey('language.pk'), nullable=False)
    parameter_pk = Column(Integer, ForeignKey('parameter.pk'), nullable=False)
    contribution_pk = Column(Integer, ForeignKey('contribution.pk'))
    source = Column(Unicode, doc='textual description of the source for the valueset')

    parameter = relationship('Parameter', innerjoin=True, backref='valuesets')

    @staticmethod
    def refine_factory_query(query):
        return query.options(
            joinedload(ValueSet.values),
            joinedload(ValueSet.parameter),
            joinedload(ValueSet.language),
        )

    @declared_attr
    def contribution(cls):
        return relationship(
            'Contribution', backref=backref('valuesets', order_by=cls.parameter_pk))

    @declared_attr
    def language(cls):
        return relationship(
            'Language', innerjoin=True, backref=backref('valuesets', order_by=cls.language_pk))

    @property
    def name(self):
        return self.language.name + ' / ' + self.parameter.name


class ValueSetReference(Base, HasSourceNotNullMixin):

    """References for a set of values (related to one parameter and one language).

    These references can be interpreted as justifications why a language does not "have"
    certain values for a parameter, too.
    """

    __table_args__ = (UniqueConstraint('valueset_pk', 'source_pk', 'description'),)

    valueset_pk = Column(Integer, ForeignKey('valueset.pk'), nullable=False)
    valueset = relationship(ValueSet, innerjoin=True, backref="references")



from clld.web.util.htmllib import *


def test_HTML():
    assert 'class' in HTML.div(HTML.cdata(), class_='abc')
    assert str(HTML.a()) == '<a></a>'


def test_literal():
    assert literal.escape('<div/>') == '&lt;div/&gt;'



"""Default DataTable for Language objects."""
from clld.web.datatables.base import DataTable, Col, LinkCol, LinkToMapCol, IdCol


class Languages(DataTable):

    """Default DataTable for Language objects."""

    def col_defs(self):
        return [
            IdCol(self, 'id'),
            LinkCol(self, 'name'),
            LinkToMapCol(self, 'm'),
            Col(self,
                'latitude',
                sDescription='<small>The geographic latitude</small>'),
            Col(self,
                'longitude',
                sDescription='<small>The geographic longitude</small>'),
        ]



from pyramid.i18n import get_localizer, TranslationStringFactory

from clld.web.util import helpers
from clld.web.assets import environment


def add_renderer_globals(module, event):
    if event['request']:
        _add_localizer(event['request'])
        event['_'] = event['request'].translate
        prod = event['request'].registry.settings.get('clld.environment') == 'production'
        environment.debug = not prod
    else:
        event['_'] = lambda s, **kw: s
    event['assets'] = environment
    event['h'] = helpers
    event['u'] = module  # pragma: no cover
    if module:
        # this is the hook for clld apps to provide additional template context for
        # templates associated with default views: provide a python with an appropriately
        # named function, and this function will be invoked, passing the current template
        # context as keyword parameters, and adding the items from the returned dict
        # to the template context.
        _renderer = event.get('renderer_name', '')
        if _renderer.endswith('.mako'):
            func = getattr(module, _renderer[:-5].replace('/', '_').replace('.', '_'), 0)
            if func:
                for k, v in func(**event).items():
                    event[k] = v


tsf = TranslationStringFactory('clld')


def add_localizer(event):
    _add_localizer(event.request)


def _add_localizer(request):
    if hasattr(request, 'translate'):
        return

    request._LOCALE_ = request.params.get('__locale__', 'en')
    localizer = get_localizer(request)

    def auto_translate(*args, **kwargs):
        return localizer.translate(tsf(*args, **kwargs))

    request._ = request.translate = auto_translate


def init_map(event):
    event.request.map = event.request.get_map()



from clld.web.util.htmllib import HTML

__all__ = ['BASE_URL', 'url', 'logo', 'link']

BASE_URL = 'https://glottolog.org'


def url(id=None, obj_type='languoid'):
    if id is None:
        return BASE_URL
    assert obj_type in ('languoid', 'reference')
    return '{0}/resource/{1}/id/{2}'.format(BASE_URL, obj_type, id)


def logo(req, width='20'):
    return HTML.img(
        src=req.static_url('clld:web/static/images/glottolog.png'),
        width=width,
        style="margin-top: -2px")


def link(req, id=None, obj_type='languoid', label=None):
    akw = dict(
        title='Glottolog' if id is None else 'View {0} at Glottolog'.format(obj_type),
        href=url(id=id, obj_type=obj_type))
    return HTML.a(logo(req), label or '', **akw)



# coding: utf8
"""Functionality to create Coins, i.e. context objects in spans.

.. seealso:: https://en.wikipedia.org/wiki/COinS
"""
import re
from urllib.parse import urlencode

from clldutils.misc import encoded, to_binary


FIELDS = {
    'any': {
        'rft_id': None,
        'aulast': None,  # First author's family name. This may be more than one word.
        # In many citations, the author's family name is recorded first and is followed
        # by a comma, i.e. Smith, Fred James is recorded as "aulast=smith"
        'aufirst': None,  # First author's given name or names or initials. This data
        # element may contain multiple words and punctuation, i.e. "Fred F", "Fred James"
        'auinit': None,  # First author's first and middle initials.
        'auinit1': None,  # First author's first initial.
        'auinitm': None,  # First author's middle initial.
        'ausuffix': None,  # First author's name suffix. Qualifiers on an author's name
        # such as "Jr.", "III" are entered here. i.e. Smith, Fred Jr. is recorded as
        # "ausuffux=jr"
        'au': None,  # This data element contains the full name of a single author, i. e.
        # "Smith, Fred M", "Harry S. Truman". (au is repeatable)
        'aucorp': None,  # Organization or corporation that is the author or creator of
        # the book, i.e. "Mellon Foundation"
        'title': None,  # Book title. Provided for compatibility with version 0.1.
        # Prefer btitle.
        'place': None,  # Place of publication. "New York"
        'pub': None,  # Publisher name. "Harper and Row"
        'date': None,  # Date of publication. Book dates are assumed to be a single year.
    },
    'dissertation': {
        'inst': None,
        'degree': None,
    },
    'book': {
        'btitle': None,  # The title of the book. This can also be expressed as title, for
        # compatibility with version 0.1. "moby dick or the white whale"
        'isbn': None,  # International Standard Book Number (ISBN). The ISBN is usually
        # presented as 9 digits plus a final check digit (which may be "X"), i.e.
        # "057117678X" but it may contain hyphens, i.e. "1-878067-73-7"
        'atitle': None,  # Chapter title. Chapter title is included if it is a distinct
        # title, i.e. "The Push Westward."
        'edition': None,  # Statement of the edition of the book. This will usually be a
        # phrase, with or without numbers, but may be a single number. I.e.
        # "First edition", "4th ed."
        'tpages': None,  # Total pages. Total pages is the largest recorded number of
        # pages, if this can be determined. I.e., "ix, 392 p." would be recorded as "392"
        # in tpages. This data element is usually available only for monographs (books and
        # printed reports). In some cases, tpages may not be numeric, i.e. "F36"
        'series': None,  # The title of a series in which the book or document was issued.
        # There may also be an ISSN associated with the series.
        'spage': None,  # First page number of a start/end (spage-epage) pair.
        # Note that pages are not always numeric.
        'epage': None,  # Second (ending) page number of a start/end (spage-epage) pair.
        'pages': None,  # Start and end pages for parts of a book, i.e. "124-147". This
        # can also be used for an unstructured pagination statement when data relating to
        # pagination cannot be interpreted as a start-end pair, i.e. "A7, C4-9", "1-3,6".
        # This data element includes the OpenURL 0.1 definition of "pages".
        'issn': None,  # International Standard Serials Number (ISSN). The issn may
        # contain a hyphen, i.e. "1041-5653". An ISSN in the book format is often
        # associated with a series title.
        'genre': lambda val: val if val in [
            'book',  # a publication that is complete in one part or a designated finite
            # number of parts, often identified with an ISBN.
            'bookitem',  # a defined section of a book, usually with a separate title or
            # number.
            'proceeding',  # a conference paper or proceeding published in a conference
            # publication.
            'conference',  # a publication bundling the proceedings of a conference.
            'report',  # report or technical report is a published document that is
            # issued by an organization, agency or government body.
            'document',  # general document type to be used when available data elements
            # do not allow determination of a more specific document type, i.e. when one
            # has only author and title but no publication information.
            'unknown',  # use when the genre of the document is unknown.
        ] else None,
    },
    'journal': {
        'rft_id': None,
        'atitle': None,  # Article title.
        'jtitle': None,  # Journal title. Use the most complete title available.
        # Abbreviated titles, when known, are records in stitle. This can also be
        # expressed as title, for compatibility with version 0.1. "journal of the
        # american medical association"
        'stitle': None,  # Abbreviated or short journal title. This is used for journal
        # title abbreviations, where known, i.e. "J Am Med Assn"
        'volume': None,  # Volume designation.Volume is usually expressed as a number but
        # could be roman numerals or non-numeric, i.e. "124", or "VI".
        'issue': None,  # This is the designation of the published issue of a journal,
        # corresponding to the actual physical piece in most cases. While usually numeric,
        # it could be non-numeric. Note that some publications use chronology in the place
        # of enumeration, i.e. Spring, 1998.
        'spage': None,  # First page number of a start/end (spage-epage) pair. Note that
        # pages are not always numeric.
        'epage': None,  # Second (ending) page number of a start/end (spage-epage) pair.
        'pages': None,  # Start and end pages, i.e. "53-58". This can also be used for an
        # unstructured pagination statement when data relating to pagination cannot be
        # interpreted as a start-end pair, i.e. "A7, C4-9", "1-3,6". This data element
        # includes the OpenURL 0.1 definition of "pages".
        'artnum': None,  # Article number assigned by the publisher. Article numbers are
        # often generated for publications that do not have usable pagination, in
        # particular electronic journal articles, i.e. "unifi000000090". A URL may be the
        # only usable identifier for an online article, in which case the URL can be
        # treated as an identifier for the article (i.e.
        # "rft_id=http://www.firstmonday.org/ issues/issue6_2/odlyzko/ index.html").
        'issn': None,  # International Standard Serials Number (ISSN). The issn may
        # contain a hyphen, i.e. "1041-5653"
        'eissn': None,  # ISSN for electronic version of the journal. Although there is no
        # distinction by format in the assignment of ISSNs, some bibliographic services
        # now carry both the ISSN for the paper version and a separate ISSN for the
        # electronic version. This data element is included here to allow the OpenURL to
        # carry both ISSNs and distinguish them.
        'isbn': None,  # International Standard Book Number (ISBN). The ISBN is usually
        # presented as 9 digits plus a final check digit (which may be "X"), i.e.
        # "057117678X" but it may contain hyphens, i.e. "1-878067-73-7"
        'coden': None,  # CODEN
        'sici': None,  # Serial Item and Contribution Identifier (SICI)
        'genre': lambda val: val if val in [
            'issue',  # one instance of the serial publication.
            'article',  # a document published in a journal.
            'proceeding',  # a single conference presentation published in a journal or
            # serial publication
            'conference',  # a record of a conference that includes one or more conference
            # papers and that is published as an issue of a journal or serial publication
            'preprint',  # an individual paper or report published in paper or
            # electronically prior to its publication in a journal or serial.
            'unknown',  # use when the genre of the document is unknown.
        ] else None,
        'chron': None,  # Enumeration or chronology in not-normalized form, i.e.
        # "1st quarter". Where numeric dates are also available, place the numeric portion
        # in the "date" Key. So a recorded date of publication of "1st quarter 1992"
        # becomes date=1992&chron=1st quarter. Normalized indications of chronology can be
        # provided in the ssn and quarter Keys.
        'ssn': lambda val: val if val in ['spring', 'summer', 'fall', 'winter'] else None,
        # Season (chronology). Legitimate values are spring, summer, fall, winter
        'quarter': lambda val: val if int(val) in [1, 2, 3, 4] else None,
        # Quarter (chronology).
        'part': None,  # Part can be a special subdivision of a volume or it can be the
        # highest level division of the journal. Parts are often designated with letters
        # or names, i.e. "B", "Supplement".
    },
}


def _encoded(value):
    if not isinstance(value, bytes) and not isinstance(value, str):
        value = '%s' % value
    return encoded(value)


class ContextObject(list):

    """A Context Object which knows how to render it's metadata as HTML span tags."""

    def __init__(self, sid, mtx, *data):
        self.sid = sid
        self.mtx = mtx
        list.__init__(self)
        for key, val in data:
            validator = FIELDS[self.mtx].get(key, FIELDS['any'].get(key)) or _encoded
            key = 'rft.' + key if not key.startswith('rft') else key
            self.append((key, validator(val)))

    @classmethod
    def from_bibtex(cls, sid, rec):
        mtx, genre = {
            'article': ('journal', 'article'),
            'book': ('book', 'book'),
            'inbook': ('book', 'bookitem'),
            'incollection': ('book', 'bookitem'),
            'inproceedings': ('book', 'proceeding'),
            'conference': ('book', 'proceeding'),
            'mastersthesis': ('dissertation', None),
            'phdthesis': ('dissertation', None),
            'proceedings': ('book', 'conference'),
            'techreport': ('book', 'report'),
            'unpublished': ('book', 'document'),
            'misc': ('book', 'unknown'),
        }.get(getattr(rec.genre, 'value', rec.genre),  # allow EnumSymbol as genre.
              ('book', 'document' if rec.get('author') else 'unknown'))
        if genre:
            data = [('genre', genre)]
        else:
            data = []

        if mtx == 'journal':
            if 'title' in rec:
                data.append(('atitle', rec['title']))
            if 'journal' in rec:
                data.append(('jtitle', rec['journal']))
        elif mtx == 'book':
            if 'title' in rec:
                data.append(('btitle' if genre == 'book' else 'atitle', rec['title']))
            if 'booktitle' in rec:
                if genre != 'book':
                    data.append(('btitle', rec['booktitle']))
        elif mtx == 'dissertation':
            if 'title' in rec:
                data.append(('title', rec['title']))
            data.append(
                ('degree', 'phd' if
                 getattr(rec.genre, 'value', rec.genre) == 'phdthesis' else 'masters'))
            data.append(('inst', rec.get('school', rec.get('institution', ''))))

        if 'url' in rec:
            data.append(('rft_id', rec['url']))

        for bibfield, openurlfield in {
            'address': 'place',
            'publisher': 'pub',
            'year': 'date',
            'volume': 'volume',
            'number': 'issue',
            'series': 'series',
            'edition': 'edition',
            'pages': 'pages',
        }.items():
            if bibfield in rec:
                data.append((openurlfield, rec[bibfield]))

        for i, author in enumerate(rec.getall('author')):
            if i == 0:
                parts = re.split(r'\s*,\s*', author, 1)
                if len(parts) == 1:
                    parts = re.split(r'\s+', author)
                    last = parts[-1]
                    first = ' '.join(parts[:-1])
                else:
                    last, first = parts
                data.append(('aulast', last))
                data.append(('aufirst', first))
            else:
                data.append(('au', author))

        return cls(sid, mtx, *data)

    def __str__(self):
        pairs = [
            (to_binary('ctx_ver'), to_binary('Z39.88-2004')),
            (to_binary('rft_val_fmt'),
             to_binary('info:ofi/fmt:kev:mtx:') + _encoded(self.mtx)),
            (to_binary('rfr_id'), to_binary('info:sid/') + _encoded(self.sid))]
        for pair in self:
            pairs.append((_encoded(pair[0]), _encoded(pair[1])))
        try:
            return urlencode(pairs)
        except UnicodeDecodeError:  # pragma: no cover
            return to_binary('')

    def span_attrs(self):
        return {'class': 'Z3988', 'title': str(self)}



import functools
import itertools
import collections

from sqlalchemy import Column, Integer, Unicode, UniqueConstraint, ForeignKey
from sqlalchemy.orm import relationship, joinedload

from zope.interface import implementer
from clldutils.color import qualitative_colors

from clld.db.meta import Base, PolymorphicBaseMixin, DBSession
from clld.web.icon import Icon
from clld import interfaces

from . import (
    IdNameDescriptionMixin,
    DataMixin, HasDataMixin, FilesMixin, HasFilesMixin,
    Language)

__all__ = ('DomainElement', 'Parameter', 'Combination')


class DomainElement_data(Base, DataMixin):
    pass


class DomainElement_files(Base, FilesMixin):
    pass


@implementer(interfaces.IDomainElement)
class DomainElement(Base,
                    PolymorphicBaseMixin,
                    IdNameDescriptionMixin,
                    HasDataMixin,
                    HasFilesMixin):

    """DomainElements can be used to model controlled lists of values for a Parameter."""

    __table_args__ = (
        UniqueConstraint('parameter_pk', 'name'),
        UniqueConstraint('parameter_pk', 'number'),
    )

    parameter_pk = Column(Integer, ForeignKey('parameter.pk'), nullable=False)

    number = Column(Integer, doc='numerical value of the domain element')
    """the number is used to sort domain elements within the domain of one parameter"""

    abbr = Column(Unicode, doc='abbreviated name')
    """abbreviated name, e.g. as label for map legends"""

    def url(self, request):
        return request.resource_url(self.parameter, _anchor='DE-' + self.id)


class Parameter_data(Base, DataMixin):
    pass


class Parameter_files(Base, FilesMixin):
    pass


@implementer(interfaces.IParameter)
class Parameter(Base,
                PolymorphicBaseMixin,
                IdNameDescriptionMixin,
                HasDataMixin,
                HasFilesMixin):

    """A measurable attribute of a language."""

    __table_args__ = (UniqueConstraint('name'),)

    domain = relationship(
        'DomainElement', backref='parameter', order_by=DomainElement.number)


class CombinationDomainElement(object):
    def __init__(self, combination, domainelements, icon=None):
        self.number = tuple(de.number for de in domainelements)
        self.id = '-'.join(str(n) for n in self.number)
        self.name = ' / '.join(de.name for de in domainelements)
        self.icon = icon
        self.languages = []
        super(CombinationDomainElement, self).__init__()


@implementer(interfaces.ICombination)
class Combination(object):

    """A combination of parameters."""

    delimiter = '_'

    def __init__(self, *parameters):
        """Initialize.

        :param parameters: distinct Parameter instances.
        """
        assert len(parameters) < 5
        assert len(set(parameters)) == len(parameters)
        self.id = self.delimiter.join(str(p.id) for p in parameters)
        self.name = ' / '.join(p.name for p in parameters)
        self.parameters = parameters
        # we keep track of languages with multiple values.
        self.multiple = []
        super(Combination, self).__init__()

    def __json__(self, *args, **kw):
        return {k: getattr(self, k) for k in ['id', 'name']}

    @classmethod
    def get(cls, id_, **kw):
        params = []
        for pid in set(id_.split(cls.delimiter)):
            params.append(
                DBSession.query(Parameter)
                .filter(Parameter.id == pid)
                .options(joinedload(Parameter.domain))
                .one())
        return cls(*params)

    @functools.cached_property
    def domain(self):
        """Compute the domain as cartesian product of constituent domains.

        .. note::

            This does only work well with parameters which have a discrete domain.
        """
        d = collections.OrderedDict()
        for i, des in enumerate(itertools.product(*[p.domain for p in self.parameters])):
            cde = CombinationDomainElement(self, des)
            d[cde.number] = cde

        for cde, color in zip(d.values(), qualitative_colors(i + 1)):
            cde.icon = Icon(color.replace('#', 'c'))

        for language, values in itertools.groupby(
            sorted(self.values, key=lambda v: v.valueset.language_pk),
            lambda i: i.valueset.language,
        ):
            # values may contain multiple values for the same parameter, so we have to
            # group those, too.
            values_by_parameter = collections.OrderedDict()
            for p in self.parameters:
                values_by_parameter[p.pk] = []
            for v in values:
                values_by_parameter[v.valueset.parameter_pk].append(v)
            for i, cv in enumerate(itertools.product(*values_by_parameter.values())):
                d[tuple(v.domainelement.number for v in cv)].languages.append(language)
                if i > 0:
                    # a language with multiple values, store a reference.
                    self.multiple.append(language)
        self.multiple = set(self.multiple)
        return list(d.values())

    @functools.cached_property
    def values(self):
        from . import ValueSet, Value

        def _filter(query, operation):
            q = query.filter(Parameter.pk == self.parameters[0].pk)
            return getattr(q, operation)(
                *[query.filter(Parameter.pk == p.pk) for p in self.parameters[1:]])

        # determine relevant languages, i.e. languages having a value for all parameters:
        languages = _filter(
            DBSession.query(Language.pk).join(ValueSet).join(Parameter),
            'intersect').subquery()

        # value query:
        return _filter(
            DBSession.query(Value)
            .join(Value.valueset)
            .join(ValueSet.parameter)
            .filter(ValueSet.language_pk.in_(languages))
            .options(
                joinedload(Value.domainelement),
                joinedload(Value.valueset).joinedload(ValueSet.language)),
            'union').all()



"""Default DataTable for Value objects."""
from sqlalchemy.orm import joinedload

from clld.db.models.common import (
    Value, ValueSet, Parameter, DomainElement, Language, Contribution, ValueSetReference,
)
from clld.db.util import icontains
from clld.web.datatables.base import (
    DataTable, LinkCol, DetailsRowLinkCol, LinkToMapCol,
)
from clld.web.datatables.base import RefsCol as BaseRefsCol
from clld.web.util.helpers import map_marker_img
from clld.web.util.htmllib import HTML, literal


class ValueNameCol(LinkCol):

    """Render the label for a Value."""

    def get_obj(self, item):
        return item.valueset

    def get_attrs(self, item):
        label = str(item)
        title = label
        if self.dt.parameter:
            label = HTML.span(map_marker_img(self.dt.req, item), literal('&nbsp;'), label)
        return {'label': label, 'title': title}

    def order(self):
        return DomainElement.number \
            if self.dt.parameter and self.dt.parameter.domain \
            else Value.description

    def search(self, qs):
        if self.dt.parameter and self.dt.parameter.domain:
            return DomainElement.name.__eq__(qs)
        return icontains(Value.description, qs)


class ValueSetCol(LinkCol):

    """Render a link to the corresponding ValueSet."""

    def get_obj(self, item):
        return item.valueset

    def get_attrs(self, item):
        return {'label': item.valueset.name}


class RefsCol(BaseRefsCol):

    """Listing sources for the corresponding ValueSet."""

    def get_obj(self, item):
        return item.valueset


class Values(DataTable):

    """Default DataTable for Value objects."""

    __constraints__ = [Parameter, Contribution, Language]

    def base_query(self, query):
        query = query.join(ValueSet).options(
            joinedload(
                Value.valueset
            ).joinedload(
                ValueSet.references
            ).joinedload(
                ValueSetReference.source
            )
        )

        if self.language:
            query = query.join(ValueSet.parameter)
            return query.filter(ValueSet.language_pk == self.language.pk)

        if self.parameter:
            query = query.join(ValueSet.language)
            query = query.outerjoin(DomainElement).options(
                joinedload(Value.domainelement))
            return query.filter(ValueSet.parameter_pk == self.parameter.pk)

        if self.contribution:
            query = query.join(ValueSet.parameter)
            return query.filter(ValueSet.contribution_pk == self.contribution.pk)

        return query

    def col_defs(self):
        #
        # TODO: move the first col def to apics-specific table!
        #
        name_col = ValueNameCol(self, 'value')
        if self.parameter and self.parameter.domain:
            name_col.choices = [de.name for de in self.parameter.domain]

        res = [DetailsRowLinkCol(self, 'd')]

        if self.parameter:
            return res + [
                LinkCol(self,
                        'language',
                        model_col=Language.name,
                        get_object=lambda i: i.valueset.language),
                name_col,
                RefsCol(self, 'source'),
                LinkToMapCol(self, 'm', get_object=lambda i: i.valueset.language),
            ]

        if self.language:
            return res + [
                name_col,
                LinkCol(self,
                        'parameter',
                        sTitle=self.req.translate('Parameter'),
                        model_col=Parameter.name,
                        get_object=lambda i: i.valueset.parameter),
                RefsCol(self, 'source'),
            ]

        return res + [
            name_col,
            ValueSetCol(self, 'valueset', bSearchable=False, bSortable=False),
            #
            # TODO: contribution col?
            #
        ]



import datetime
import collections

from sqlalchemy import (
    Column, String, Unicode, Date, Integer, ForeignKey, Boolean,
    UniqueConstraint)
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.declarative import declared_attr

from zope.interface import implementer

from clld.db.meta import Base, PolymorphicBaseMixin, DBSession
from clld import interfaces
from clld.web.util.htmllib import HTML

from . import (
    IdNameDescriptionMixin,
    DataMixin, HasDataMixin, FilesMixin, HasFilesMixin,
    Contributor)

__all__ = ('Dataset', 'Editor')


class Dataset_data(Base, DataMixin):

    """Associated data mapper."""


class Dataset_files(Base, FilesMixin):

    """Associated files mapper."""


@implementer(interfaces.IDataset)
class Dataset(Base,
              PolymorphicBaseMixin,
              IdNameDescriptionMixin,
              HasDataMixin,
              HasFilesMixin):

    """Represents a database.

    Each project (e.g. WALS, APiCS) is regarded as one dataset; thus, each app will
    have exactly one Dataset object.
    """

    published = Column(Date, default=datetime.date.today, doc='date of publication')
    publisher_name = Column(Unicode, doc='publisher')
    publisher_place = Column(Unicode, doc='place of publication')
    publisher_url = Column(String)
    license = Column(String, default="http://creativecommons.org/licenses/by/3.0/")
    domain = Column(String, nullable=False)
    contact = Column(String)

    def get_stats(self, resources, **filters):
        res = collections.OrderedDict()
        for rsc in resources:
            if rsc.name != 'combination':
                query = DBSession.query(rsc.model)
                if rsc.name in filters:
                    query = query.filter(filters[rsc.name])
                res[rsc.name] = query.count()
        return res

    def formatted_editors(self):
        def _format(eds):
            return ' & '.join(ed.contributor.last_first() for ed in eds)

        res = _format([e for e in self.editors if e.primary])
        secondary = [e for e in self.editors if not e.primary]
        if secondary:
            res = ' with '.join([res, _format(secondary)])  # pragma: no cover
        return res

    def formatted_name(self):
        return HTML.span(
            self.name,
            **{
                'xmlns:dct': "http://purl.org/dc/terms/",
                'href': "http://purl.org/dc/dcmitype/Dataset",
                'property': "dct:title",
                'rel': "dct:type",
                'class': 'Dataset'}
        )


class Editor(Base, PolymorphicBaseMixin):

    """Many-to-many association between contributors and dataset."""

    __table_args__ = (UniqueConstraint('dataset_pk', 'contributor_pk'),)

    dataset_pk = Column(Integer, ForeignKey('dataset.pk'), nullable=False)
    contributor_pk = Column(Integer, ForeignKey('contributor.pk'), nullable=False)

    # contributors are ordered.
    ord = Column(Integer, default=1)

    # we distinguish between primary and secondary (a.k.a. 'with ...') contributors.
    primary = Column(Boolean, default=True)

    contributor = relationship(
        Contributor, innerjoin=True, lazy=False, backref=backref('editor', uselist=False))

    @declared_attr
    def dataset(cls):
        return relationship(
            Dataset, innerjoin=True, backref=backref(
                'editors', order_by=[cls.primary.desc(), cls.ord], lazy=False))



import pytest
from clld.db.models import common
from clld.web.datatables.source import Sources

from testutils import handle_dt


@pytest.mark.parametrize(
    "params",
    [
        {},
        {
            'language': 'language',
            'sSearch_5': 'book',
            'iSortingCols': '1',
            'iSortCol_0': '5',
            'sSortDir_0': 'desc'},
    ])
def test_Sources(request_factory, params):
    with request_factory(params=params) as req:
        handle_dt(req, Sources, common.Source)



from sqlalchemy import Column, Integer, UniqueConstraint, ForeignKey
from sqlalchemy.orm import relationship

from clld.db.meta import Base

from . import IdNameDescriptionMixin, Language

__all__ = ('GlossAbbreviation',)


class GlossAbbreviation(Base, IdNameDescriptionMixin):

    """Possibly language-specific abbreviation used in IGTs."""

    __table_args__ = (UniqueConstraint('id', 'language_pk'),)

    language_pk = Column(Integer, ForeignKey('language.pk'))
    language = relationship(Language, backref="gloss_abbreviations")



import json

import pytest

from clld.db.models.common import Parameter, Language
from clld.web.adapters import geojson
from clld.web.datatables.base import DataTable

geojson.pacific_centered()


def test_GeoJson(mocker):
    adapter = geojson.GeoJson(None)
    assert len(list(adapter.feature_iterator(None, None))) == 0
    assert len(list(adapter.feature_iterator(Language(), None))) == 1
    assert len(list(adapter.feature_iterator(mocker.Mock(languages=[Language()]), None))) == 1


def test_GeoJsonParameter(env, request_factory):
    adapter = geojson.GeoJsonParameter(None)
    assert '{' in adapter.render(Parameter.get('no-domain'), env['request'])

    with request_factory(params=dict(domainelement='de')) as req:
        res = json.loads(adapter.render(Parameter.get('parameter'), req))
        assert len(res['features']) > 0
        assert 'label' in res['features'][0]['properties']


def test_GeoJsonParameterMultipleValueSets(env):
    adapter = geojson.GeoJsonParameterMultipleValueSets(None)
    assert '{' in adapter.render(Parameter.get('no-domain'), env['request'])


def test_GeoJsonParameterFlatProperties(env):
    adapter = geojson.GeoJsonParameterFlatProperties(None)
    assert '{' in adapter.render(Parameter.get('no-domain'), env['request'])


def test_GeoJsonLanguages(env):
    class MockLanguages(DataTable):
        def get_query(self, *args, **kw):
            return [Language.first()]

    adapter = geojson.GeoJsonLanguages(None)
    assert 'Point' in\
        adapter.render(
            MockLanguages(env['request'], Language), env['request'])


def test_get_lonlat(mocker):
    assert geojson.get_lonlat(None) is None
    assert geojson.get_lonlat((None, 5)) is None
    assert geojson.get_lonlat((-50, 1))[0] > 0
    assert geojson.get_lonlat(mocker.Mock(latitude=1, longitude=1)) == (pytest.approx(1), pytest.approx(1))


def test_get_feature(data):
    l = Language.first()
    assert geojson.get_feature(l)['id'] == l.id
    assert geojson.get_feature(l)['properties']['name'] == l.name
    assert geojson.get_feature(l, name='geo')['properties']['name'] == 'geo'



import pytest
from clld.db.models import common
from clld.web.datatables import Values

from testutils import handle_dt


@pytest.mark.parametrize(
    "params",
    [
        {},
        {
            'sSearch_0': '> 1',
            'sSearch_1': 's',
            'iSortingCols': '3',
            'iSortCol_0': '0',
            'sSortDir_0': 'desc',
            'iSortCol_1': '1',
            'sSortDir_1': 'desc',
            'iSortCol_2': '2',
            'sSortDir_2': 'desc',
        },
        {'contribution': 'contribution'},
        {'language': 'language'},
        {'parameter': 'parameter'},
        {'parameter': 'parameter', 'sSearch_2': 's'},
        {'parameter': 'no-domain'},
    ])
def test_Values(request_factory, params):
    with request_factory(params=params) as req:
        handle_dt(req, Values, common.Value)



"""Functionality to represent clld objects as RDF graphs and serializations."""
from sqlalchemy.orm import load_only
from clldutils.misc import xmlchars

from clld.web.adapters.base import Representation, Index
from clld.lib.rdf import convert


class Rdf(Representation):

    """Virtual base class."""

    rdflibname = None

    def render(self, ctx, req):
        return convert(
            xmlchars(super(Rdf, self).render(ctx, req)), 'xml', self.rdflibname)


class RdfIndex(Index):

    """Basically a rdf sitemap.

    .. note::

        To make this reasonably performant even for large collections of resources,
        we only pass resource ids and names to the template. This may still be not enough -
        think Glottolog refs - in which case the DataTable should implement a suitable
        `rdf_index_query`, e.g. with a LIMIT clause.
    """

    rdflibname = None

    def render(self, ctx, req):
        if req.params.get('sEcho'):
            # triggered from a datatable, thus potentially filtered and sorted
            items = ctx.get_query(limit=1000)
        else:
            # triggered without any filter parameters
            items = ctx.rdf_index_query(req.db.query(ctx.db_model()).order_by(ctx.db_model().pk))
        if isinstance(ctx.model.name, property):
            items = [(item.id, None) for item in items.options(load_only('id'))]
        else:
            items = [(item.id, item.name)
                     for item in items.options(load_only('id', 'name'))]
        return convert(super(RdfIndex, self).render(items, req), 'xml', self.rdflibname)



"""
This module provides base classes to compose DataTables.

I.e. objects which have a double
nature: On the client they provide the information to instantiate a jquery DataTables
object. Server side they know how to provide the data to the client-side table.
"""
import re
import typing
import functools

from markupsafe import Markup
from sqlalchemy.orm import undefer
from sqlalchemy.types import String, Unicode, Float, Integer, Boolean
from zope.interface import implementer, implementedBy
from clldutils.misc import nfilter
from pyramid.request import Request

from clld.db.meta import DBSession, Base
from clld.db.util import icontains, as_int
from clld.web.util.htmllib import HTML, literal
from clld.web.util.helpers import (
    link, button, icon, JS_CLLD, external_link, linked_references, JSDataTable,
)
from clld.web.util.component import Component
from clld.interfaces import IDataTable, IIndex


OPERATOR_PATTERN = re.compile(r'\s*(?P<op>>=?|<=?|==?)\s*')

DISPLAY_LENGTH, DISPLAY_LIMIT = 100, 1000


def type_coerce(type_, value, fallback):
    """Return type_(value), on ValueError return fallback."""
    # turn user parameters into int/float where there is no try/except already
    try:
        return type_(value)
    except ValueError:
        return fallback


def filter_number(col, qs, type_=None, qs_weight=1):
    """Extended filter operators for numbers.

    :param col: an sqlalchemy column instance.
    :param qs: a string, providing a filter criterion.
    :return: sqlalchemy filter expression.
    """
    if '..' in qs:
        min_, max_ = qs.split('..', 1)
        # We only support a single range!
        max_ = max_.split('..')[0]
        return (
            filter_number(col, '>=' + min_, type_=type_, qs_weight=qs_weight),
            filter_number(col, '<=' + max_, type_=type_, qs_weight=qs_weight))
    op = col.__eq__
    match = OPERATOR_PATTERN.match(qs)
    if match:
        op = {
            '>': col.__gt__,
            '>=': col.__ge__,
            '=': col.__eq__,
            '==': col.__eq__,
            '<': col.__lt__,
            '<=': col.__le__,
        }.get(match.group('op'), col.__eq__)
        qs = qs[match.end():]
    try:
        if type_:
            qs = type_(qs.strip())
        else:
            if isinstance(col.property.columns[0].type, Float):
                qs = float(qs.strip()) * qs_weight
            if isinstance(col.property.columns[0].type, Integer):
                qs = int(qs.strip()) * qs_weight
        return op(qs)
    except ValueError:  # pragma: no cover
        # if we cannot form a proper filter argument, we return None
        return


class Col(object):

    """DataTables are basically a list of column specifications.

    A column in a DataTable typically corresponds to a column of an sqlalchemy model.
    This column can either be supplied directly via a model_col keyword argument, or we
    try to look it up as attribute with name "name" on self.dt.model.
    """

    dt_name_pattern = re.compile('[a-z]+[A-Z]+[a-z]+')

    # convenient way to provide defaults for some kw arguments of __init__:
    __kw__ = {}

    def __init__(self, dt, name, get_object=None, model_col=None, format=None, **kw):
        """
        :param kw: Camel-cased keywords with type prefix are made available to JS. Recognized \
        keywords are \
        - `sDescription`: Description of the column, possibly using HTML,\
        - `bSearchable`\
        - `bSortable`
        -
        """
        self.dt = dt
        self.name = name
        self._get_object = get_object
        self._format = format
        self.model_col = model_col
        self.model_col_type = None
        self.js_args = {
            'sName': name,
            'sTitle': '' if not name
            else self.dt.req.translate(name.replace('_', ' ').capitalize())}

        # take the defaults into account:
        for k, v in self.__kw__.items():
            kw.setdefault(k, v)

        for key, val in kw.items():
            if self.dt_name_pattern.match(key):
                self.js_args[key] = val
            else:
                setattr(self, key, val)

        if not self.model_col:
            #
            # model_col was not explicitly passed as keyword parameter
            #
            # TODO: fix mechanism to infer model_col for derived classes!)
            #
            model_col = getattr(self.dt.model, self.name, None)
            if model_col and hasattr(model_col.property, 'columns'):
                self.model_col = model_col

        if self.model_col:
            self.model_col_type = self.model_col.property.columns[0].type

        if isinstance(self.model_col_type, Boolean):
            if not hasattr(self, 'choices'):
                self.choices = ['True', 'False']
            if not hasattr(self, 'input_size'):
                self.input_size = 'small'
        elif isinstance(self.model_col_type, (Float, Integer)):
            self.js_args.setdefault('sClass', 'right')
            if not hasattr(self, 'input_size'):
                self.input_size = 'small'

    def get_obj(self, item):
        """Get the object for formatting and filtering.

        .. note::

            derived columns with a model_col not on self.dt.model should override this
            method.
        """
        if getattr(self, '_get_object'):
            return self._get_object(item)
        return item

    def get_value(self, item):
        mc = self.model_col
        val = getattr(self.get_obj(item), mc.name if mc else self.name, None)
        return '' if val is None else val

    def format_value(self, value):
        if isinstance(self.model_col_type, Boolean):
            return '%s' % value
        if isinstance(self.model_col_type, Float) and isinstance(value, float):
            return ('%.' + str(getattr(self, 'precision', 2)) + 'f') % value
        return value

    #
    # external API called by DataTable objects:
    #
    def order(self):
        """Called when collecting the order by clauses of a datatable's search query."""
        return self.model_col

    def search(self, qs):
        """Called when collecting the filter criteria of a datatable's search query."""
        if isinstance(self.model_col_type, (String, Unicode)):
            if getattr(self, 'choices', None):
                # make sure select box values match sharp!
                return self.model_col.__eq__(qs)
            else:
                return icontains(self.model_col, qs)
        if isinstance(self.model_col_type, (Float, Integer)):
            return filter_number(self.model_col, qs)
        if isinstance(self.model_col_type, Boolean):
            return self.model_col.__eq__(qs == 'True')

    def format(self, item):
        """Called when converting the matching result items of a datatable to json."""
        if getattr(self, '_format'):
            return self._format(item)
        return self.format_value(self.get_value(item))


class ExternalLinkCol(Col):

    """Format item as external link."""

    __kw__ = {'bSearchable': False, 'bSortable': False}

    def get_attrs(self, item):
        return {}

    def format(self, item):
        url = getattr(self.get_obj(item), 'url', None)
        return external_link(url, **self.get_attrs(item)) if url else ''


class PercentCol(Col):

    """Treat a model col of type float as percentage."""

    def search(self, qs):
        return filter_number(self.model_col, qs, qs_weight=0.01)

    def format_value(self, value):
        return '%.0f%%' % (100 * value,)


class LinkCol(Col):

    """Column which renders a link."""

    def get_attrs(self, item):
        return {}

    def format(self, item):
        obj = self.get_obj(item)
        return link(self.dt.req, obj, **self.get_attrs(item)) if obj else ''


class IdCol(LinkCol):

    """Render an ID."""

    __kw__ = {'sClass': 'right', 'input_size': 'mini'}

    def get_attrs(self, item):
        return {'label': self.get_obj(item).id}

    def search(self, qs):
        if self.model_col:
            return self.model_col.__eq__(qs)


class IntegerIdCol(IdCol):

    """Render a numeric ID."""

    __kw__ = {'input_size': 'mini', 'sClass': 'right', 'sTitle': 'No.'}

    def search(self, qs):
        return filter_number(as_int(self.model_col), qs, type_=int)

    def order(self):
        return as_int(self.model_col)


class LinkToMapCol(Col):

    """Render a button.

    We use the CLLD.Map.showInfoWindow API function to construct a button to open
    a popup on the map.
    """

    __kw__ = {'bSearchable': False, 'bSortable': False, 'sTitle': '', 'map_id': 'map'}

    def format(self, item):
        obj = self.get_obj(item)
        if not obj or getattr(obj, 'latitude', None) is None:
            return ''
        return HTML.a(
            icon('icon-globe'),
            title='show %s on map' % getattr(obj, 'name', ''),
            href="#" + self.map_id,
            onclick=JS_CLLD.mapShowInfoWindow(self.map_id, obj.id),
            class_='btn',
        )


class DetailsRowLinkCol(Col):

    """Render a button to open a details row on the fly.

    .. seealso:: http://www.datatables.net/examples/api/row_details.html
    """

    __kw__ = {
        'bSearchable': False,
        'bSortable': False,
        'sClass': 'center',
        'sType': 'html',
        'sTitle': 'Details',
        'button_text': 'more',
    }

    def format(self, item):
        return button(
            self.button_text,
            href=self.dt.req.resource_url(self.get_obj(item), ext='snippet.html'),
            title="show details",
            class_="btn-info details",
            tag=HTML.button)


class RefsCol(Col):

    """Column listing linked sources."""

    __kw__ = dict(bSearchable=False, bSortable=False)

    def format(self, item):
        vs = self.get_obj(item)
        return ', '.join(
            nfilter([getattr(vs, 'source', None), linked_references(self.dt.req, vs)]))


class Toolbar(Component):
    def __init__(self, req, ctx, obj, dl_url_tmpl, interface, **kw):
        self.req = req
        self.ctx = ctx
        self.obj = obj
        self.dl_url_tmpl = dl_url_tmpl
        self.interface = interface
        self.dl_formats = kw.pop('dl_formats', {})
        kw.setdefault('doc_position', 'right')
        kw.setdefault('exclude', ['html', 'snippet.html'])
        self._kw = kw
        self._opener_class = 'dl-widget-%s' % self.interface.__name__
        self._id_prefix = 'dt-dl-' if IDataTable.providedBy(ctx) else 'rsc-dl-'

    def get_options(self):
        return {'doc_position': 'left'}

    def get_default_options(self):
        return self._kw

    def js(self):
        return Markup(HTML.script(literal("""\
    $(document).ready(function() {
        $('.%s').clickover({
            html: true,
            title: 'Column information',
            placement: '%s',
            trigger: 'click'
        });
    });""" % (self._opener_class, self.options['doc_position']))))

    def doc(self):
        items = []
        for col in self.ctx.cols:
            dsc = col.js_args.get('sDescription')
            if dsc:
                items.extend([HTML.dt(col.js_args['sTitle']), HTML.dd(literal(dsc))])
        return HTML.dl(
            HTML.p(
                'Columns containing numeric data may be filtered giving an upper and/or lower '
                'bound in the form "<5" or a range in the form "-2..20".'),
            self.ctx.options.get('sDescription', ''),
            *items)

    def render(self, no_js=False):
        doc = HTML.div(self.doc())
        buttons = [HTML.button(
            HTML.i(class_='icon-info-sign icon-white'),
            class_='btn btn-info %s' % self._opener_class,
            **{'data-content': str(doc), 'type': 'button'})]
        for ext, label in self.dl_formats.items():
            buttons.append(HTML.button(
                label,
                class_='btn',
                id='{}-{}-download'.format(self.ctx.eid, ext),
                href='#',
                tite="Download items in the table formatted as {}.".format(label),
                onclick="document.location.href = {}; "
                "return false;".format(self.dl_url_tmpl % ext)
            ))
        res = HTML.div(*buttons, **{'class': 'btn-group' if len(buttons) > 1 else ''})
        if no_js:
            return res
        return HTML.div(res, self.js())  # pragma: no cover


@implementer(IDataTable)
class DataTable(Component):

    """
    DataTables are used to manage selections of instances of one model class.

    Often datatables are used to display only a pre-filtered set of items which are
    related to some other entity in the system. This scenario is supported as follows:
    For each model class listed
    in :py:attr:`clld.web.datatables.base.DataTable.__constraints__` an appropriate object
    specified either by keyword parameter or as request parameter will be looked up at
    datatable initialization, and placed into a datatable attribute named after the model
    class in lowercase. These attributes will be used when creating the URL for the data
    request, to make sure the same pre-filtering is applied.

    .. note::

        The actual filtering has to be done in a custom implementation
        of :py:meth:`clld.web.datatables.base.DataTable.base_query`.

    Items in a table may have standard representations other than tabular data - e.g. sources
    may be more useful as BibTeX records. If so, and the custom serialization is registered as
    adapter for the model class, DataTables can provide access to the currently filtered items
    in this custom format through a download button. This can be configured by adding a key
    `dl_formats` to :attr:`clld.web.datatables.base.DataTable.__toolbar_kw__` specifying a `dict`
    mapping registered `extension` strings to labels for the download button.
    """

    __template__ = 'clld:web/templates/datatable.mako'
    __constraints__ = []
    __toolbar_kw__ = {}

    def __init__(
            self, req: Request, model: typing.Type[Base], eid: typing.Optional[str] = None, **kw):
        """
        :param req: request object.
        :param model: mapper class, instances of this class will be the rows in the table.
        :param eid: HTML element id that will be assigned to this data table.
        """
        self.model = model
        self.req = req
        self.eid = eid or self.__class__.__name__
        self.count_all = None
        self.count_filtered = None
        self.filters = []
        self._toolbar = Toolbar(
            req,
            self,
            model(),
            JSDataTable.current_url(self.eid, '%s'),
            IIndex,
            **self.__toolbar_kw__ or {})

        for _model in self.__constraints__:
            attr = self.attr_from_constraint(_model)
            if kw.get(attr):
                setattr(self, attr, kw[attr])
            elif attr in req.params:
                setattr(self, attr, _model.get(req.params[attr], default=None))
            else:
                setattr(self, attr, None)

    @staticmethod
    def attr_from_constraint(model):
        return model.__name__.lower()

    def __str__(self):
        return '%ss' % self.model.__name__

    def __repr__(self):
        return '%ss' % self.model.__name__

    def col_defs(self):
        """Must be implemented by derived classes.

        :return: list of instances of :py:class:`clld.web.datatables.base.Col`.
        """
        return [LinkCol(self, 'name')]

    @functools.cached_property
    def cols(self):
        return self.col_defs()

    def xhr_query(self):
        """Get additional URL parameters for XHR.

        :return: a mapping to be passed as query parameters to the server when requesting\
        table data via xhr.
        """
        res = {}
        for _model in self.__constraints__:
            attr = self.attr_from_constraint(_model)
            if getattr(self, attr):
                res[attr] = getattr(self, attr).id
        return res

    def get_default_options(self):
        query_params = {}
        query_params.update(self.req.query_params)
        query_params.update(self.xhr_query() or {})
        try:
            data_url = self.req.route_url(
                '%ss' % self.model.__name__.lower(), _query=query_params)
        except KeyError:
            interface = list(implementedBy(self.model))[0]
            data_url = self.req.route_url(
                '%ss' % interface.__name__.lower()[1:], _query=query_params)
        return {
            "language": {
                "paginate": {
                    "next": self.req.translate("Next"),
                    "previous": self.req.translate("Previous"),
                    "first": self.req.translate("First"),
                    "last": self.req.translate("Last"),
                },
                "emptyTable": self.req.translate("No data available in table"),
                "info": self.req.translate("Showing _START_ to _END_ of _TOTAL_ entries"),
                "infoEmpty": self.req.translate("Showing 0 to 0 of 0 entries"),
                "infoFiltered": self.req.translate("(filtered from _MAX_ total entries)"),
                "lengthMenu": self.req.translate("Show _MENU_ entries"),
                "loadingRecords": self.req.translate("Loading..."),
                "processing": self.req.translate("Processing..."),
                "search": self.req.translate("Search:"),
                "zeroRecords": self.req.translate("No matching records found"),
            },
            'bServerSide': True,
            'bProcessing': True,
            "sDom": "<'dt-before-table row-fluid'<'span4'i><'span6'p><'span2'f<'"
            + self.eid + "-toolbar'>>r>t<'span4'i><'span6'p>",
            "bAutoWidth": False,
            "sPaginationType": "bootstrap",
            "aoColumns": [col.js_args for col in self.cols],
            "iDisplayLength": DISPLAY_LENGTH,
            "aLengthMenu": [[50, 100, 200], [50, 100, 200]],
            'sAjaxSource': data_url,
        }

    def db_model(self):
        return self.model

    def base_query(self, query):
        """Custom DataTables can overwrite this method to add joins, or apply filters.

        :return: ``sqlalchemy.orm.query.Query`` instance.
        """
        return query

    def rdf_index_query(self, query):
        """Custom DataTables can overwrite this method to apply filters. They should
        however, not add any performance overhead like joins or joinedload options.
        """
        return query

    def default_order(self):
        return self.db_model().pk

    def get_query(self, limit=DISPLAY_LIMIT, offset=0, undefer_cols=()):
        query = self.base_query(
            DBSession.query(self.db_model()).filter(self.db_model().active == True))
        self.count_all = query.count()

        _filters = []
        for name, val in self.req.params.items():
            if val and name.startswith('sSearch_'):
                try:
                    colindex = int(name.split('_')[1])
                    col = self.cols[colindex]
                    clauses = col.search(val)
                except (ValueError, IndexError):  # pragma: no cover
                    clauses = None
                if clauses is not None:
                    if not isinstance(clauses, (tuple, list)):
                        clauses = [clauses]
                    for clause in clauses:
                        if clause is not None:
                            query = query.filter(clause)
                            _filters.append((colindex, col.js_args['sTitle'], val))
        for colindex, coltitle, qs in sorted(set(_filters)):
            self.filters.append((coltitle, qs))

        self.count_filtered = query.count()

        iSortingCols = type_coerce(int, self.req.params.get('iSortingCols', 0), 0)

        for index in range(min(iSortingCols, 10)):
            try:
                col = self.cols[int(self.req.params.get('iSortCol_%s' % index))]
            except (TypeError, ValueError, IndexError):  # pragma: no cover
                continue
            if col.js_args.get('bSortable', True):
                orders = col.order()
                if orders is not None:
                    if not isinstance(orders, (tuple, list)):
                        orders = [orders]
                    for order in orders:
                        if self.req.params.get('sSortDir_%s' % index) == 'desc':
                            order = order.desc()
                        query = query.order_by(order)

        clauses = self.default_order()
        if not isinstance(clauses, (list, tuple)):
            clauses = (clauses,)
        query = query.order_by(*clauses)

        if 'iDisplayLength' in self.req.params:
            limit = type_coerce(int, self.req.params['iDisplayLength'], DISPLAY_LENGTH)
            # make sure no more than DISPLAY_LIMIT items can be selected
            limit = min(limit, DISPLAY_LIMIT)
        query = query\
            .limit(DISPLAY_LIMIT if limit == -1 else limit)\
            .offset(type_coerce(int, self.req.params.get('iDisplayStart', offset), offset))

        if undefer_cols:
            query = query.options(*(undefer(c) for c in undefer_cols))

        return query

    def render(self):
        return Component.render(self) + self._toolbar.js()

    def toolbar(self):
        return self._toolbar.render(no_js=True)



import pytest

from clld import RESOURCES


def _xml_findall(app, name):
    return list(app.parsed_body.findall('.//%s' % name))


def test_robots(app):
    res = app.get('/robots.txt')
    assert 'User-agent: robot1\nDisallow: /\n' in res.body.decode('utf8')


def test_sitemapindex(app):
    app.get_xml('/sitemap.xml')
    assert len(app.parsed_body.findall(
        '{http://www.sitemaps.org/schemas/sitemap/0.9}sitemap')) > 0


def test_sitemap(app):
    app.get_xml('/sitemap.language.0.xml')
    app.get_json('/resourcemap.json?rsc=language')
    app.get_json('/resourcemap.json?rsc=parameter')
    app.get_json('/resourcemap.json?rsc=xxx', status=404)


def test_dataset(app):
    _ = app.get('/void.md.bib')
    res = app.get_html('/?__admin__=1')
    assert 'notexisting.css' in res
    assert 'notexisting.js' in res
    # Test content negotiation:
    app.get('/', accept='image/png', status=406)
    app.get_xml('/', accept='application/rdf+xml')
    assert app.parsed_body.find(
        './/{http://rdfs.org/ns/void#}Dataset').get(
        '{http://www.w3.org/1999/02/22-rdf-syntax-ns#}about')
    assert 'skos:example' in app.get_xml('/void.rdf')
    # Test __locale__ param is propagated:
    res = app.get_html('/?__locale__=en')
    assert '?__locale__=en' in res


@pytest.mark.parametrize('rsc', [rsc for rsc in RESOURCES if rsc.with_index])
def test_resources(app, rsc):
    app.get_html('/{0}s/{0}'.format(rsc.name))
    app.get_html('/{0}s/{0}.snippet.html'.format(rsc.name), docroot='div')
    res = app.get_xml('/{0}s/{0}.rdf'.format(rsc.name))
    assert res.headers['Vary'] == 'Accept'
    assert len(_xml_findall(app, '{http://www.w3.org/2004/02/skos/core#}scopeNote')) == 1
    assert len(_xml_findall(app, '{http://www.w3.org/2004/02/skos/core#}altLabel')) > 0
    app.get_html('/%ss' % rsc.name)
    app.get_xml('/%ss.rdf' % rsc.name)
    app.get_dt('/%ss?iDisplayLength=5' % rsc.name)


def test_resources_special_cases(app):
    app.get_xml('/unitparameters/up2.rdf')
    app.get_html('/combinations/parameter')
    app.get_xml('/combinations/parameter.rdf')


def test_source(app):
    app.get('/sources/source.bib')
    app.get('/sources.bib')
    app.get_xml('/sources.rdf?sEcho=1')
    # resources with a name should be listed with rdfs:label in rdf index.
    # see https://github.com/clld/clld/issues/66
    assert len(app.parsed_body.findall(
        '{http://www.w3.org/1999/02/22-rdf-syntax-ns#}Description')) > 0


@pytest.mark.parametrize(
    'route,status',
    [('/languages/replaced', 301), ('/languages/gone', 410), ('/sources/replaced', 301)]
)
def test_replacement(app, route, status):
    app.get(route, status=status)



"""
Functionality to handle bibligraphical data in the BibTeX format.

.. seealso:: http://en.wikipedia.org/wiki/BibTeX
"""
import re
import pathlib
import collections

from zope.interface import Interface, implementer
from clldutils.misc import to_binary
from clldutils.source import Source

from clld.util import DeclEnum
from clld.lib import latex

__all__ = [
    'u_unescape', 'unescape', 'stripctrlchars',
    'EntryType', 'FIELDS', 'IRecord', 'Record', 'IDatabase', 'Database']

latex.register()

UU_PATTERN = re.compile(r'\?\[\\u(?P<number>[0-9]{3,5})]')


def u_unescape(s):
    """Unencode Unicode escape sequences.

    Match all 3-5-digit sequences with unicode character
    replace all '?[\\u....]' with corresponding unicode

    There are some decimal/octal mismatches in unicode encodings in bibtex
    """
    def iterchunks(s):
        end = 0
        for m in UU_PATTERN.finditer(s):
            yield s[end:m.start()]
            yield chr(int(m.group('number')))
            end = m.end()
        yield s[end:]

    return ''.join(iterchunks(s))


SYMBOLS = {
    '\\plusminus{}': '\xb1',
    '\\middot{}': '\xb7',
    '\\textopeno{}': "\u0254",
    '\\dh{}': "\u00f0",
    '\\DH{}': "\u00d0",
    '\\textthorn{}': "\u00fe",
    '\\textless{}': "<",
    '\\textgreater{}': ">",
    '\\circ{}': "\u00b0",
    '\\textltailn{}': "\u0272",
    '\\textlambda{}': "\u03BB",
    '\\textepsilon{}': '\u025b',
    '\\textquestiondown{}': '\xbf',
    '\\textschwa{}': '\u0259',
    '\\textsubdot{o}': '\u1ecd',
    '\\textrhooktopd{}': '\u0257',
    # '\\eurosign{}': '\u20ac',
    '\\eurosign{}': '\u2021',
    '\\textquestiondown': '\xbf',
    '\\textquotedblleft': '\u201c',
    '\\textquotedblright': '\u201d',
    '\\textquoteleft': '\u2018',
    '\\textquoteright': '\u2019',

    '\\textsubdot{D}': '\u1e0c',
    '\\textsubdot{E}': '\u1eb8',
    '\\textsubdot{H}': '\u1e24',
    '\\textsubdot{I}': '\u1eca',
    '\\textsubdot{O}': '\u1ecc',
    '\\textsubdot{T}': '\u1e6c',
    '\\textsubdot{d}': '\u1e0d',
    '\\textsubdot{b}': '\u1e05',
    '\\textsubdot{e}': '\u1eb9',
    '\\textsubdot{h}': '\u1e25',
    '\\textsubdot{i}': '\u1ecb',
    '\\textsubdot{n}': '\u1e47',
    '\\textsubdot{r}': '\u1e5b',
    '\\textsubdot{s}': '\u1e63',
    '\\textsubdot{t}': '\u1e6d',
    '\\ng{}': '\u014b',
    '\\oslash{}': '\u00f8',
    '\\Oslash{}': '\u00d8',
    '\\textdoublebarpipe{}': '\u01c2',
    # '\\dots': '',
    '\\Aa{}': '\xc5',
    '\\Aa{}Rsj\xd6': '\xc5rsj\xf6',

    '\\guillemotleft': '\xab',
    '\\guillemotleft{}': '\xab',
    '\\guillemotright': '\xbb',
}

RE_XML_ILLEGAL = re.compile(
    r'([\u0000-\u0008\u000b-\u000c\u000e-\u001f\ufffe-\uffff])'
    + r'|'
    + r'([%s-%s][^%s-%s])|([^%s-%s][%s-%s])|([%s-%s]$)|(^[%s-%s])' %
    (
        chr(0xd800), chr(0xdbff), chr(0xdc00), chr(0xdfff),
        chr(0xd800), chr(0xdbff), chr(0xdc00), chr(0xdfff),
        chr(0xd800), chr(0xdbff), chr(0xdc00), chr(0xdfff),
    ))


def stripctrlchars(string):
    try:
        return RE_XML_ILLEGAL.sub("", string)
    except TypeError:
        return string


def unescape(string):
    r"""Transform latex escape sequences of type \`\ae  into unicode.

    :param string: str or binary

    :return: str
    """
    def _delatex(s):
        try:
            result = to_binary(s, encoding='latin1').decode('latex+latin1')
        except UnicodeEncodeError:
            result = string
        return str(result)

    res = u_unescape(stripctrlchars(_delatex(string.strip())))
    for symbol in sorted(SYMBOLS.keys(), key=lambda s: len(s)):
        res = res.replace(symbol, SYMBOLS[symbol])
    if '\\' not in res:
        res = res.replace('{', '').replace('}', '')
    return res


class EntryType(DeclEnum):

    """Bibtext entry types.

    article
        An article from a journal or magazine.
        Required fields: author, title, journal, year
        Optional fields: volume, number, pages, month, note, key

    book
        A book with an explicit publisher.
        Required fields: author/editor, title, publisher, year
        Optional fields: volume/number, series, address, edition, month, note, key

    booklet
        A work that is printed and bound, but without a named publisher or sponsoring
        institution.
        Required fields: title
        Optional fields: author, howpublished, address, month, year, note, key

    conference
        The same as inproceedings, included for Scribe compatibility.

    inbook
        A part of a book, usually untitled. May be a chapter (or section or whatever)
        and/or a range of pages.
        Required fields: author/editor, title, chapter/pages, publisher, year
        Optional fields: volume/number, series, type, address, edition, month, note, key

    incollection
        A part of a book having its own title.
        Required fields: author, title, booktitle, publisher, year
        Optional fields: editor, volume/number, series, type, chapter, pages, address,
        edition, month, note, key

    inproceedings
        An article in a conference proceedings.
        Required fields: author, title, booktitle, year
        Optional fields: editor, volume/number, series, pages, address, month,
        organization, publisher, note, key

    manual
        Technical documentation.
        Required fields: title
        Optional fields: author, organization, address, edition, month, year, note, key

    mastersthesis
        A Master's thesis.
        Required fields: author, title, school, year
        Optional fields: type, address, month, note, key

    misc
        For use when nothing else fits.
        Required fields: none
        Optional fields: author, title, howpublished, month, year, note, key

    phdthesis
        A Ph.D. thesis.
        Required fields: author, title, school, year
        Optional fields: type, address, month, note, key

    proceedings
        The proceedings of a conference.
        Required fields: title, year
        Optional fields: editor, volume/number, series, address, month, publisher,
        organization, note, key

    techreport
        A report published by a school or other institution, usually numbered within a
        series.
        Required fields: author, title, institution, year
        Optional fields: type, number, address, month, note, key

    unpublished
        A document having an author and title, but not formally published.
        Required fields: author, title, note
        Optional fields: month, year, key
    """

    article = 'article', 'article'  # Article
    book = 'book', 'book'  # Book
    booklet = 'booklet', 'booklet'
    conference = 'conference', 'conference'  # Conference
    inbook = 'inbook', 'inbook'  # BookSection
    incollection = 'incollection', 'incollection'
    inproceedings = 'inproceedings', 'inproceedings'
    manual = 'manual', 'manual'  # Manual
    mastersthesis = 'mastersthesis', 'mastersthesis'  # Thesis
    misc = 'misc', 'misc'
    phdthesis = 'phdthesis', 'phdthesis'  # Thesis
    proceedings = 'proceedings', 'proceedings'  # Proceedings
    techreport = 'techreport', 'techreport'  # Report
    unpublished = 'unpublished', 'unpublished'  # Manuscript


FIELDS = [
    'address',  # Publisher's address
    'annote',  # An annotation for annotated bibliography styles (not typical)
    'author',  # The name(s) of the author(s) (separated by and)
    'booktitle',  # The title of the book, if only part of it is being cited
    'chapter',  # The chapter number
    'crossref',  # The key of the cross-referenced entry
    'edition',  # The edition of a book, long form (such as "First" or "Second")
    'editor',  # The name(s) of the editor(s)
    'eprint',  # A specification of electronic publication, preprint or technical report
    'howpublished',  # How it was published, if the publishing method is nonstandard
    'institution',  # institution involved in the publishing,not necessarily the publisher
    'journal',  # The journal or magazine the work was published in
    'key',  # A hidden field used for specifying or overriding the orderalphabetical order
    'month',  # The month of publication (or, if unpublished, the month of creation)
    'note',  # Miscellaneous extra information
    'number',  # The "(issue) number" of a journal, magazine, or tech-report
    'organization',  # The conference sponsor
    'pages',  # Page numbers, separated either by commas or double-hyphens.
    'publisher',  # The publisher's name
    'school',  # The school where the thesis was written
    'series',  # The series of books the book was published in
    'title',  # The title of the work
    'type',  # The field overriding the default type of publication
    'url',  # The WWW address
    'volume',  # The volume of a journal or multi-volume book
    'year',
]

FIELDS_SET = set(FIELDS)


class _Convertable(object):

    """Mixin adding a shortcut to clld.lib.bibutils.convert as method."""

    def format(self, fmt):
        if fmt == 'txt':
            if hasattr(self, 'text'):
                return self.text()
            raise NotImplementedError('no text method found!')
        return str(self)


class IRecord(Interface):

    """marker."""


@implementer(IRecord)
class Record(Source, _Convertable):

    """A BibTeX record is an ordered dict with two special properties - id and genre.

    To overcome the limitation of single values per field in BibTeX, we allow fields,
    i.e. values of the dict to be iterables of strings as well.
    Note that to support this use case comprehensively, various methods of retrieving
    values will behave differently. I.e. values will be

    - joined to a string in __getitem__,
    - retrievable as assigned with get (i.e. only use get if you know how a value was\
      assigned),
    - retrievable as list with getall

    .. note:: Unknown genres are converted to "misc".
    """

    def __init__(self, genre, id_, *args, **kw):
        super(Record, self).__init__(genre, 'a', args, **kw)
        self.id = id_
        if isinstance(self.genre, str):
            try:
                self.genre = EntryType.from_string(genre.lower())
            except ValueError:
                self.genre = EntryType.misc

    @classmethod
    def from_object(cls, obj, **kw):
        data = {field: getattr(obj, field, None) for field in FIELDS if getattr(obj, field, None)}
        data.update(kw)
        data.setdefault('title', obj.description)
        data = sorted(data.items())
        return cls(obj.bibtex_type, obj.id, *data)

    @classmethod
    def from_string(cls, bibtexString, lowercase=False):
        return cls.from_bibtex(bibtexString, lowercase=lowercase)

    @staticmethod
    def sep(key):
        return ' and ' if key in ('author', 'editor') else '; '

    def getall(self, key):
        """Get list of all values for key.

        :return: list of strings representing the values of the record for field 'key'.
        """
        res = self.get(key, [])
        if isinstance(res, str):
            res = res.split(Record.sep(key))
        return [_f for _f in res if _f]

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key):
        """Get concatenated string of all values for key.

        :return: string representing the concatenation of the values for field 'key'.
        """
        value = collections.OrderedDict.__getitem__(self, key)
        if not isinstance(value, (tuple, list)):
            value = [value]
        return Record.sep(key).join(filter(None, value))

    def __str__(self):
        """Represent the record in BibTeX format.

        :return: string encoding the record in BibTeX syntax.
        """
        keys = sorted(self, key=lambda k: (k not in FIELDS_SET, k))
        m = max([0] + list(map(len, keys)))

        fields = ("  %s = {%s}" % (k.ljust(m), self[k]) for k in keys)

        return "@%s{%s,\n%s\n}" % (
            getattr(self.genre, 'value', self.genre), self.id, ",\n".join(fields))


class IDatabase(Interface):

    """marker."""


@implementer(IDatabase)
class Database(_Convertable):

    """Represents a bibtex databases, i.e. a container class for Record instances."""

    def __init__(self, records):
        self.records = [r for r in records if r and r.genre and r.id]
        self._keymap = None

    def __str__(self):
        return '\n'.join(str(r) for r in self.records)

    @property
    def keymap(self):
        """Map bibtex record ids to list index."""
        if self._keymap is None:
            self._keymap = dict((r.id, i) for i, r in enumerate(self.records))
        return self._keymap

    @classmethod
    def from_file(cls, bibFile, encoding='utf8', lowercase=False):
        """Create bibtex database from a bib-file.

        @param bibFile: path of the bibtex-database-file to be read.
        """
        bibFile = pathlib.Path(bibFile)
        content = bibFile.read_text(encoding=encoding) if bibFile.exists() else ''

        return cls((Record.from_string('@' + m, lowercase=lowercase)
                    for m in re.split(r'^\s*@', content, 0, re.MULTILINE)))

    def __len__(self):
        return len(self.records)

    def __getitem__(self, key):
        """Access bib records by index or citation key."""
        return self.records[key if isinstance(key, int) else self.keymap[key]]

    def __iter__(self):
        return iter(self.records)



import pathlib

from sqlalchemy import Column, Integer, String, Unicode, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr

__all__ = (
    'IdNameDescriptionMixin',
    'FilesMixin', 'HasFilesMixin',
    'DataMixin', 'HasDataMixin',
)


class IdNameDescriptionMixin(object):

    """Mixin for 'visible' objects, i.e. anything that has to be displayed.

    In particular all :doc:`resources` fall into this category.

    .. note::

        Only one of :py:attr:`clld.db.models.common.IdNameDescriptionMixin.description`
        or :py:attr:`clld.db.models.common.IdNameDescriptionMixin.markup_description`
        should be supplied, since these are used mutually exclusively.
    """

    id = Column(String, unique=True)
    """A ``str`` identifier of an object which can be used for sorting and as part of a
    URL path; thus should be limited to characters valid in URLs, and should not contain
    '.' or '/' since this may trip up route matching.
    """

    name = Column(Unicode)
    """A human readable 'identifier' of the object."""

    description = Column(Unicode)
    """A description of the object."""

    markup_description = Column(Unicode)
    """A description of the object containing HTML markup."""


# ----------------------------------------------------------------------------
# We augment mapper classes for basic objects using mixins to add the ability
# to store arbitrary key-value pairs and files associated with an object.
# ----------------------------------------------------------------------------
class FilesMixin(IdNameDescriptionMixin):

    """This mixin provides a way to associate files with instances of another model class.

    .. note::

        The file itself is not stored in the database but must be created in the
        filesystem, e.g. using the create method.
    """

    @classmethod
    def owner_class(cls):
        return cls.__name__.split('_')[0]

    ord = Column(Integer, default=1)
    """Ordinal to control sorting of files associated with one db object."""

    mime_type = Column(String)
    """Mime-type of the file content."""

    @declared_attr
    def object_pk(cls):
        return Column(Integer, ForeignKey('%s.pk' % cls.owner_class().lower()))

    @property
    def relpath(self):
        """OS file path of the file relative to the application's file-system dir."""
        return pathlib.Path(self.owner_class().lower()) / str(self.object.id) / str(self.id)

    def create(self, dir_, content):
        """Write ``content`` to a file using ``dir_`` as file-system directory.

        :return: File-system path of the file that was created.
        """
        p = pathlib.Path(dir_).joinpath(self.relpath)
        if not p.parent.exists():
            p.parent.mkdir(parents=True)
        p.write_bytes(content.encode('utf8') if isinstance(content, str) else content)
        return str(p)


class HasFilesMixin(object):

    """Mixin for model classes which may have associated files."""

    @property
    def files(self):
        """return ``dict`` of associated files keyed by ``id``."""
        return dict((f.id, f) for f in self._files)

    @declared_attr
    def _files(cls):
        return relationship(cls.__name__ + '_files', backref='object')


class DataMixin(object):

    """Provide a simple way to attach key-value pairs to a model class given by name."""

    @classmethod
    def owner_class(cls):
        return cls.__name__.split('_')[0]

    key = Column(Unicode)
    value = Column(Unicode)
    ord = Column(Integer, default=1)

    @declared_attr
    def object_pk(cls):
        return Column(Integer, ForeignKey('%s.pk' % cls.owner_class().lower()))


class HasDataMixin(object):

    """Adds a convenience method to retrieve the key-value pairs from data as dict.

    .. note::

        It is the responsibility of the programmer to make sure conversion to a ``dict``
        makes sense, i.e. the keys in data are actually unique, thus usable as dictionary
        keys.
    """

    def datadict(self):
        """return ``dict`` of associated key-value pairs."""
        return dict((d.key, d.value) for d in self.data)

    @declared_attr
    def data(cls):
        return relationship(cls.__name__ + '_data', order_by=cls.__name__ + '_data.ord')



from pathlib import Path

from clld.db.meta import DBSession
from clld.db.models import common


def test_Config_replacement_key():
    assert common.Config.replacement_key('X', 'Y') == '__X_Y__'
    assert common.Config.replacement_key(None, 'Y') == '__NoneType_Y__'


def test_Files(db, tmp_path, persist):
    l = common.Sentence(id='abc', name='Name')
    f = common.Sentence_files(object=l, id='abstract', mime_type='audio/mpeg')
    p = f.create(tmp_path, 'content')
    assert Path(p).exists()

    l._files.append(f)
    persist(l)
    DBSession.refresh(l)
    assert l.files
    assert l.audio


def test_Dataset(db, persist):
    from clld import RESOURCES

    d = persist(common.Dataset(id='abc', domain='test'))
    assert d.jsondata == d.jsondatadict
    d.get_stats(RESOURCES, source=common.Source.id == None)


def test_Contributor():
    d = common.Contributor(id='abc')
    d.last_first()
    d = common.Contributor(id='abc', name='Robert Forkel')
    assert d.last_first() == 'Forkel, Robert'
    d = common.Contributor(id='abc', name='Hans Robert von Forkel')
    assert d.last_first() == 'von Forkel, Hans Robert'


def test_Language():
    d = common.Language(id='abc')
    assert d.glottocode is None
    assert d.iso_code is None


def test_Source():
    d = common.Source(id='abc')
    assert d.gbs_identifier is None
    d = common.Source(id='abc', jsondata={'gbs': {'volumeInfo': {}}})
    assert d.gbs_identifier is None
    d = common.Source(
        id='abc',
        jsondata={
            'gbs': {
                'volumeInfo': {
                    'industryIdentifiers': [{'type': 'x', 'identifier': 'y'}]}}})
    assert d.gbs_identifier == 'y'
    d = common.Source(
        id='abc',
        jsondata={
            'gbs': {
                'volumeInfo': {
                    'industryIdentifiers': [{'type': 'ISBN_10', 'identifier': ''}]}}})
    assert d.gbs_identifier == 'ISBN:'
    d = common.Source(
        id='abc',
        jsondata={
            'gbs': {
                'volumeInfo': {
                    'industryIdentifiers': [{'type': 'ISBN_13', 'identifier': ''}]}}})
    assert d.gbs_identifier == 'ISBN:'
    d.bibtex()


def test_Data(db, persist):
    l = common.Language(id='abc', name='Name')
    l.data.append(common.Language_data(key='abstract', value='c'))
    persist(l)
    DBSession.refresh(l)
    assert l.datadict()['abstract'] == 'c'


def test_Unit():
    _ = common.Unit(name='unit', language=common.Language(name='language'))


def test_UnitValue(db, persist):
    from clld.db.models.common import Unit, Language, UnitParameter, UnitValue, UnitDomainElement

    u = Unit(name='unit', language=Language(name='language'))
    p1 = UnitParameter()
    p2 = UnitParameter()
    # NOTE: we assume paramter of UnitValue and UnitDomainElement are identical
    #       (i.e. we do not enforce/check this)
    v = UnitValue(
        unit=u, unitparameter=p1,
        unitdomainelement=UnitDomainElement(parameter=p1, name='ude'))
    assert str(v) == 'ude'
    persist(v, p2)
    try:
        v.unitparameter_pk = p2.pk
        raise ValueError  # pragma: no cover
    except AssertionError:
        pass

    v.unitparameter_pk = p1.pk
    DBSession.flush()


def test_unitvalue_to_string():
    de_name = common.UnitDomainElement(id='de_name', name='Code with name')
    de_noname = common.UnitDomainElement(id='de_noname')
    val_de_name = common.UnitValue(id='val_de_name', unitdomainelement=de_name)
    val_de_noname = common.UnitValue(id='val_de_noname', unitdomainelement=de_noname)
    val_name = common.UnitValue(id='val_name', name='Value with name')
    val_noname = common.UnitValue(id='val_noname')

    assert str(val_de_name) == 'Code with name'
    assert str(val_de_noname) == 'val_de_noname'
    assert str(val_name) == 'Value with name'
    assert str(val_noname) == 'val_noname'


def test_Identifier():
    i = common.Identifier(id='a', name='a', type=common.IdentifierType.iso.value)
    assert i.url()
    i = common.Identifier(id='a', name='a', type='xxx')
    assert i.url() is None


def test_Contribution(data):
    c = DBSession.query(common.Contribution).first()
    assert c.formatted_contributors()


def test_Value(data):
    assert 'valueset' in common.Value.first().__json__(None)


def test_value_to_string():
    de_name = common.DomainElement(id='de_name', name='Code with name')
    de_noname = common.DomainElement(id='de_noname')
    val_de_name = common.Value(id='val_de_name', domainelement=de_name)
    val_de_noname = common.Value(id='val_de_noname', domainelement=de_noname)
    val_name = common.Value(id='val_name', name='Value with name')
    val_noname = common.Value(id='val_noname')

    assert str(val_de_name) == 'Code with name'
    assert str(val_de_noname) == 'val_de_noname'
    assert str(val_name) == 'Value with name'
    assert str(val_noname) == 'val_noname'


def test_Combination(data):
    p = common.Parameter.first()
    c = common.Combination.get(common.Combination.delimiter.join(2 * [p.id]))
    assert c.values
    assert c.domain
    assert c.__json__()



import time


def test_ui(selenium):
    selenium.browser.get(selenium.url('/'))
{% if cookiecutter.cldf_module %}
    time.sleep(1)
    selenium.browser.get(selenium.url('/languages'))
    time.sleep(3)
    selenium.browser.get(selenium.url('/sources'))
{% endif %}
    time.sleep(3)



"""Generic utility functions."""
import re
import random
import string
import pathlib
import contextlib

from sqlalchemy.sql.sqltypes import SchemaType
from sqlalchemy.types import TypeDecorator, Enum
from clldutils.path import move
from clldutils.declenum import DeclEnum as BaseEnum
from clldutils.lgr import ABBRS as LGR_ABBRS
assert LGR_ABBRS

__all__ = ['random_string', 'safe_overwrite', 'summary', 'DeclEnum', 'DeclEnumType']


def random_string(length):
    return ''.join(random.choice(string.ascii_lowercase) for _ in range(length))


@contextlib.contextmanager
def safe_overwrite(fname):
    fname = pathlib.Path(fname)
    if not fname.parent.exists():
        fname.parent.mkdir()
    assert fname.parent.exists()
    tmp = fname.parent
    while tmp.exists():
        tmp = fname.parent.joinpath('%s.%s' % (fname.name, random_string(6)))
    yield tmp
    if fname.exists():
        fname.unlink()
    move(tmp, fname)


def summary(text, max_length=70):
    from textwrap import shorten
    if max_length < 4:
        return '...'
    return shorten(text, width=max_length, placeholder=' ...')


# From "The Enum Recipe": http://techspot.zzzeek.org/2011/01/14/the-enum-recipe/
class DeclEnum(BaseEnum):
    @classmethod
    def db_type(cls):
        return DeclEnumType(cls)


class DeclEnumType(SchemaType, TypeDecorator):
    cache_ok = True

    def __init__(self, enum):
        self.enum = enum
        self.impl = Enum(
            *enum.values(),
            name="ck%s" % re.sub(
                '([A-Z])', lambda m: "_" + m.group(1).lower(), enum.__name__))

    def _set_table(self, table, column):
        self.impl._set_table(table, column)  # pragma: no cover

    def copy(self):
        return DeclEnumType(self.enum)  # pragma: no cover

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return value.value

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return self.enum.from_string(value.strip())



from clld.db.models import common
from clld.interfaces import ILinkAttrs, IFrequencyMarker, IDownload
from clld.web.adapters.download import N3Dump


def test_get_url_template(env):
    from clld.web.util.helpers import get_url_template

    assert get_url_template(env['request'], 'parameter', variable_map={'id': 'ID'}) == '/parameters/{ID}'


def test_newline2br(env):
    from clld.web.util.helpers import newline2br

    assert newline2br(None) == ''


def test_charis_font_spec_css(env):
    from clld.web.util.helpers import charis_font_spec_css

    assert charis_font_spec_css()


def test_collapsed(env):
    from clld.web.util.helpers import collapsed

    collapsed('eid', 'some content')


def test_glottolog_url(env):
    from clld.web.util.helpers import glottolog_url

    glottolog_url('abcd1234')


def test_maybe_license_link(env):
    from clld.web.util.helpers import maybe_license_link

    for i, text in enumerate([
        'no license url',
        'http://example.org/',
        'http://creativecommons.org/licenses/nope/4.0',
        'http://creativecommons.org/licenses/by/4.0',
        'http://en.wikipedia.org/wiki/Public_domain',
    ]):
        maybe_license_link(
            env['request'],
            text,
            button='small' if i else 'none')


def test_get_valueset(request_factory):
    from clld.web.util.helpers import get_valueset

    with request_factory(params=dict(parameter='1')) as req:
        get_valueset(req, common.Language.first())


def test_alt_representations(env):
    from clld.web.util.helpers import alt_representations

    alt_representations(env['request'], common.Parameter.first(), exclude=['xslt'])


def test_icon(env):
    from clld.web.util.helpers import icon

    assert 'white' in icon('download', inverted=True)


def test_link(env, utility_factory, mocker):
    from clld.web.util.helpers import link

    link(env['request'], common.Language(id='id', name='Name'))
    link(env['request'], common.Value.first(), class_='right')

    with utility_factory(mocker.Mock(return_value={}), ILinkAttrs):
        link(env['request'], common.Value.first())


def test_text_citation(env):
    from clld.web.util.helpers import text_citation

    text_citation(env['request'], common.Contribution.first())


def test_get_referents(env):
    from clld.web.util.helpers import get_referents

    get_referents(common.Source.first(), exclude=['language'])


def test_data_uri(env):
    from clld.web.util.helpers import data_uri

    res = data_uri(__file__, 'text/plain')
    assert not res.split(',')[1].startswith("b'")


def test_gbs_link(env, mocker):
    from clld.web.util.helpers import gbs_link

    gbs_link(mocker.MagicMock(), pages='34')
    gbs_link(mocker.MagicMock(jsondata=dict(
        gbs=dict(accessInfo=dict(viewability='NO_PAGES')))))


def test_map_marker_img(env, mocker):
    from clld.web.util.helpers import map_marker_img

    map_marker_img(env['request'], None, marker=mocker.Mock(return_value=None))


def test_maybe_external_link(env):
    from clld.web.util.helpers import maybe_external_link

    assert 'href' not in maybe_external_link('not a URL')
    assert 'href' in maybe_external_link('http://wals.info')


def test_external_link(env):
    from clld.web.util.helpers import external_link

    assert 'http://wals.info' in external_link('http://wals.info')


def test_button(env):
    from clld.web.util.helpers import button

    assert 'click' in button('click')


def test_linked_contributors(env):
    from clld.web.util.helpers import linked_contributors

    linked_contributors(env['request'], common.Contribution.first())


def test_urlescape(env):
    from clld.web.util.helpers import urlescape

    urlescape('a b c\\/ab')


def test_coins(env):
    from clld.web.util.helpers import coins

    coins(env['request'], common.Contribution.first())
    coins(env['request'], None)


def test_format_gbs_identifier(env):
    from clld.web.util.helpers import format_gbs_identifier

    format_gbs_identifier(common.Source.first())


def test_linked_references(env, mocker):
    from clld.web.util.helpers import linked_references

    assert linked_references(env['request'], None) == ''
    mocker.patch('clld.web.util.helpers.link')
    linked_references(
        env['request'],
        mocker.Mock(references=[
            mocker.MagicMock(description=''), mocker.MagicMock(description='')]))


def test_text2html(env):
    from clld.web.util.helpers import text2html

    assert '<br' in text2html('abc\ndef')
    assert 'div' in str(text2html('chunk', mode='p'))


def test_partitioned(env):
    from clld.web.util.helpers import partitioned

    assert list(partitioned(range(10)))[0] == [0, 1, 2, 3]


def test_contactmail(env):
    from clld.web.util.helpers import contactmail

    contactmail(env['request'])


def test_format_frequency(env, mocker, utility_factory):
    from clld.web.util.helpers import format_frequency

    format_frequency(env['request'], common.Value.first())
    format_frequency(env['request'], mocker.Mock(frequency=None))

    with utility_factory(mocker.Mock(return_value='url'), IFrequencyMarker):
        format_frequency(env['request'], common.Value.first())


def test_format_coordinates(env, mocker):
    from clld.web.util.helpers import format_coordinates

    r = str(format_coordinates(mocker.Mock(latitude=5.333333333333, longitude=-9.999)))
    assert "5°20" in r
    assert "10°W" in r
    format_coordinates(mocker.Mock(latitude=5.333, longitude=-9.99), no_seconds=False)
    assert format_coordinates(common.Language.get('l2')) == ''
    assert format_coordinates(common.Language.get('language')) != ''


def test_get_downloads(env, utility_factory):
    from clld.web.util.helpers import get_rdf_dumps, get_downloads

    with utility_factory(N3Dump(common.Language, 'clld'), IDownload):
        assert list(get_rdf_dumps(env['request'], common.Language))
        assert list(get_downloads(env['request']))


def test_rendered_sentence(env, mocker):
    from clld.web.util.helpers import rendered_sentence

    rendered_sentence(mocker.MagicMock())
    rendered_sentence(common.Sentence.first())
    rendered_sentence(common.Sentence.first(), abbrs=dict(SG='singular'))
    assert '1SG' in str(rendered_sentence(common.Sentence.first(), abbrs=dict()))


def test_language_identifier(env):
    from clld.web.util.helpers import language_identifier

    assert language_identifier(None, None) == ''

    lang = common.Language.get('language')
    for identifier in lang.identifiers:
        language_identifier(env['request'], identifier)


def test_localize_url(env):
    from clld.web.util.helpers import localize_url

    assert '__locale__' not in localize_url(env['request'], 'en')
    assert '__locale__=eo' in localize_url(env['request'], 'eo')
    assert '__locale__' not in localize_url(env['request'], 'eo', 'eo')



"""
Character translation utilities for LaTeX-formatted text.

Usage:
 - unicode(string,'latex')
 - ustring.decode('latex')
are both available just by letting "import latex" find this file.
 - unicode(string,'latex+latin1')
 - ustring.decode('latex+latin1')
where latin1 can be replaced by any other known encoding, also
become available by calling latex.register().

We also make public a dictionary latex_equivalents,
mapping ord(unicode char) to LaTeX code.

D. Eppstein, October 2003.
"""
import re
import sys
import codecs


def register():
    """Enable encodings of the form 'latex+x' where x describes another encoding.

    Unicode characters are translated to or from x when possible, otherwise
    expanded to latex.
    """
    codecs.register(_registry)


def getregentry():
    """Encoding module API."""
    return _registry('latex')  # pragma: no cover


def _registry(encoding):
    if encoding == 'latex':
        encoding = None  # pragma: no cover
    elif encoding.startswith('latex+'):
        encoding = encoding[6:]  # pragma: no cover
    elif sys.version_info >= (3, 9, 0) and encoding.startswith('latex_'):  # pragma: no cover
        encoding = encoding[6:]
    else:
        return None  # pragma: no cover

    class Codec(codecs.Codec):
        def encode(self, input, errors='strict'):  # pragma: no cover
            """Convert unicode string to latex."""
            output = []
            for c in input:
                if encoding:
                    try:
                        output.append(c.encode(encoding))
                        continue
                    except Exception:
                        pass
                if ord(c) in latex_equivalents:
                    output.append(latex_equivalents[ord(c)])
                else:
                    output += ['{\\char', str(ord(c)), '}']
            return ''.join(output), len(input)

        def decode(self, input, errors='strict'):
            """Convert latex source string to unicode."""
            if encoding:
                input = str(input, encoding, errors)

            # Note: we may get buffer objects here.
            # It is not permussable to call join on buffer objects
            # but we can make them joinable by calling unicode.
            # This should always be safe since we are supposed
            # to be producing unicode output anyway.
            x = map(str, _unlatex(input))
            return ''.join(x), len(input)

    class StreamWriter(Codec, codecs.StreamWriter):
        pass

    class StreamReader(Codec, codecs.StreamReader):
        pass

    return (Codec().encode, Codec().decode, StreamReader, StreamWriter)


def _tokenize(tex):  # pragma: no cover
    """Convert latex source into sequence of single-token substrings."""
    start = 0
    try:
        # skip quickly across boring stuff
        pos = next(_stoppers.finditer(tex)).span()[0]
    except StopIteration:
        yield tex
        return

    while 1:
        if pos > start:
            yield tex[start:pos]
            if (
                tex[start] == '\\'
                and not (tex[pos - 1].isdigit() and tex[start + 1].isalpha())
            ):
                while pos < len(tex) and tex[pos].isspace():  # skip blanks after csname
                    pos += 1

        while pos < len(tex) and tex[pos] in _ignore:  # pragma: no cover
            pos += 1    # flush control characters
        if pos >= len(tex):
            return
        start = pos
        if tex[pos:pos + 2] in {'$$': None, '/~': None}:  # protect ~ in urls
            pos += 2  # pragma: no cover
        elif tex[pos].isdigit():
            while pos < len(tex) and tex[pos].isdigit():
                pos += 1
        elif tex[pos] == '-':
            while pos < len(tex) and tex[pos] == '-':
                pos += 1
        elif tex[pos] != '\\' or pos == len(tex) - 1:
            pos += 1
        elif not tex[pos + 1].isalpha():
            pos += 2
        else:
            pos += 1
            while pos < len(tex) and tex[pos].isalpha():
                pos += 1
            if tex[start:pos] == '\\char' or tex[start:pos] == '\\accent':
                while pos < len(tex) and tex[pos].isdigit():  # pragma: no cover
                    pos += 1


class _unlatex(object):  # pragma: no cover

    """Convert tokenized tex into sequence of unicode strings.  Helper for decode()."""

    def __iter__(self):
        """Turn self into an iterator.  It already is one, nothing to do."""
        return self

    def __init__(self, tex):
        """Create a new token converter from a string."""
        self.tex = tuple(_tokenize(tex))  # turn tokens into indexable list
        self.pos = 0                    # index of first unprocessed token
        self.lastoutput = 'x'           # lastoutput must always be nonempty string

    def __getitem__(self, n):
        """Return token at offset n from current pos."""
        return self.pos + n < len(self.tex) and self.tex[self.pos + n] or None

    def __next__(self):
        """Find and return another piece of converted output."""
        if self.pos >= len(self.tex):
            raise StopIteration
        nextoutput = self.chunk()
        if (
            self.lastoutput[0] == '\\'
            and self.lastoutput[-1].isalpha()
            and nextoutput[0].isalpha()
        ):
            # add extra space to terminate csname
            nextoutput = ' ' + nextoutput  # pragma: no cover
        self.lastoutput = nextoutput
        return nextoutput

    def chunk(self):
        """Grab another set of input tokens and convert them to an output string."""
        for delta, c in self.candidates(0):  # pragma: no cover
            if c in _l2u:
                self.pos += delta
                return chr(_l2u[c])
            elif len(c) == 2 and c[1] == 'i' and (c[0], '\\i') in _l2u:
                self.pos += delta       # correct failure to undot i
                return chr(_l2u[(c[0], '\\i')])
            elif len(c) == 1 and c[0].startswith('\\char') and c[0][5:].isdigit():
                self.pos += delta
                return chr(int(c[0][5:]))

        # nothing matches, just pass through token as-is
        self.pos += 1
        return self[-1]

    def candidates(self, offset):
        """Generate pairs delta,c.

        Where c is a token or tuple of tokens from tex
        (after deleting extraneous brackets starting at pos) and delta
        is the length of the tokens prior to bracket deletion.
        """
        t = self[offset]
        if t in _blacklist:
            return
        elif t == '{':
            for delta, c in self.candidates(offset + 1):
                if self[offset + delta + 1] == '}':
                    yield delta + 2, c
        elif t == '\\mbox':
            for delta, c in self.candidates(offset + 1):  # pragma: no cover
                yield delta + 1, c
        elif t == '$' and self[offset + 2] == '$':
            yield 3, (t, self[offset + 1], t)  # pragma: no cover
        else:
            q = self[offset + 1]
            if q == '{' and self[offset + 3] == '}':
                yield 4, (t, self[offset + 2])
            elif q:
                yield 2, (t, q)
            yield 1, t


latex_equivalents = {
    0x0009: ' ',
    0x000a: '\n',
    0x0023: r'{\#}',
    0x0026: r'{\&}',
    0x00a0: '{~}',
    0x00a1: '{!`}',
    0x00a2: '{\\not{c}}',
    0x00a3: '{\\pounds}',
    0x00a7: '{\\S}',
    0x00a8: '{\\"{}}',
    0x00a9: '{\\copyright}',
    0x00af: '{\\={}}',
    0x00ac: '{\\neg}',
    0x00ad: '{\\-}',
    0x00b0: '{\\mbox{$^\\circ$}}',
    0x00b1: '{\\mbox{$\\pm$}}',
    0x00b2: '{\\mbox{$^2$}}',
    0x00b3: '{\\mbox{$^3$}}',
    0x00b4: "{\\'{}}",
    0x00b5: '{\\mbox{$\\mu$}}',
    0x00b6: '{\\P}',
    0x00b7: '{\\mbox{$\\cdot$}}',
    0x00b8: '{\\c{}}',
    0x00b9: '{\\mbox{$^1$}}',
    0x00bf: '{?`}',
    0x00c0: '{\\`A}',
    0x00c1: "{\\'A}",
    0x00c2: '{\\^A}',
    0x00c3: '{\\~A}',
    0x00c4: '{\\"A}',
    0x00c5: '{\\AA}',
    0x00c6: '{\\AE}',
    0x00c7: '{\\c{C}}',
    0x00c8: '{\\`E}',
    0x00c9: "{\\'E}",
    0x00ca: '{\\^E}',
    0x00cb: '{\\"E}',
    0x00cc: '{\\`I}',
    0x00cd: "{\\'I}",
    0x00ce: '{\\^I}',
    0x00cf: '{\\"I}',
    0x00d1: '{\\~N}',
    0x00d2: '{\\`O}',
    0x00d3: "{\\'O}",
    0x00d4: '{\\^O}',
    0x00d5: '{\\~O}',
    0x00d6: '{\\"O}',
    0x00d7: '{\\mbox{$\\times$}}',
    0x00d8: '{\\O}',
    0x00d9: '{\\`U}',
    0x00da: "{\\'U}",
    0x00db: '{\\^U}',
    0x00dc: '{\\"U}',
    0x00dd: "{\\'Y}",
    0x00df: '{\\ss}',
    0x00e0: '{\\`a}',
    0x00e1: "{\\'a}",
    0x00e2: '{\\^a}',
    0x00e3: '{\\~a}',
    0x00e4: '{\\"a}',
    0x00e5: '{\\aa}',
    0x00e6: '{\\ae}',
    0x00e7: '{\\c{c}}',
    0x00e8: '{\\`e}',
    0x00e9: "{\\'e}",
    0x00ea: '{\\^e}',
    0x00eb: '{\\"e}',
    0x00ec: '{\\`\\i}',
    0x00ed: "{\\'\\i}",
    0x00ee: '{\\^\\i}',
    0x00ef: '{\\"\\i}',
    0x00f1: '{\\~n}',
    0x00f2: '{\\`o}',
    0x00f3: "{\\'o}",
    0x00f4: '{\\^o}',
    0x00f5: '{\\~o}',
    0x00f6: '{\\"o}',
    0x00f7: '{\\mbox{$\\div$}}',
    0x00f8: '{\\o}',
    0x00f9: '{\\`u}',
    0x00fa: "{\\'u}",
    0x00fb: '{\\^u}',
    0x00fc: '{\\"u}',
    0x00fd: "{\\'y}",
    0x00ff: '{\\"y}',

    0x0100: '{\\=A}',
    0x0101: '{\\=a}',
    0x0102: '{\\u{A}}',
    0x0103: '{\\u{a}}',
    0x0104: '{\\c{A}}',
    0x0105: '{\\c{a}}',
    0x0106: "{\\'C}",
    0x0107: "{\\'c}",
    0x0108: "{\\^C}",
    0x0109: "{\\^c}",
    0x010a: "{\\.C}",
    0x010b: "{\\.c}",
    0x010c: "{\\v{C}}",
    0x010d: "{\\v{c}}",
    0x010e: "{\\v{D}}",
    0x010f: "{\\v{d}}",
    0x0112: '{\\=E}',
    0x0113: '{\\=e}',
    0x0114: '{\\u{E}}',
    0x0115: '{\\u{e}}',
    0x0116: '{\\.E}',
    0x0117: '{\\.e}',
    0x0118: '{\\c{E}}',
    0x0119: '{\\c{e}}',
    0x011a: "{\\v{E}}",
    0x011b: "{\\v{e}}",
    0x011c: '{\\^G}',
    0x011d: '{\\^g}',
    0x011e: '{\\u{G}}',
    0x011f: '{\\u{g}}',
    0x0120: '{\\.G}',
    0x0121: '{\\.g}',
    0x0122: '{\\c{G}}',
    0x0123: '{\\c{g}}',
    0x0124: '{\\^H}',
    0x0125: '{\\^h}',
    0x0128: '{\\~I}',
    0x0129: '{\\~\\i}',
    0x012a: '{\\=I}',
    0x012b: '{\\=\\i}',
    0x012c: '{\\u{I}}',
    0x012d: '{\\u\\i}',
    0x012e: '{\\c{I}}',
    0x012f: '{\\c{i}}',
    0x0130: '{\\.I}',
    0x0131: '{\\i}',
    0x0132: '{IJ}',
    0x0133: '{ij}',
    0x0134: '{\\^J}',
    0x0135: '{\\^\\j}',
    0x0136: '{\\c{K}}',
    0x0137: '{\\c{k}}',
    0x0139: "{\\'L}",
    0x013a: "{\\'l}",
    0x013b: "{\\c{L}}",
    0x013c: "{\\c{l}}",
    0x013d: "{\\v{L}}",
    0x013e: "{\\v{l}}",
    0x0141: '{\\L}',
    0x0142: '{\\l}',
    0x0143: "{\\'N}",
    0x0144: "{\\'n}",
    0x0145: "{\\c{N}}",
    0x0146: "{\\c{n}}",
    0x0147: "{\\v{N}}",
    0x0148: "{\\v{n}}",
    0x014c: '{\\=O}',
    0x014d: '{\\=o}',
    0x014e: '{\\u{O}}',
    0x014f: '{\\u{o}}',
    0x0150: '{\\H{O}}',
    0x0151: '{\\H{o}}',
    0x0152: '{\\OE}',
    0x0153: '{\\oe}',
    0x0154: "{\\'R}",
    0x0155: "{\\'r}",
    0x0156: "{\\c{R}}",
    0x0157: "{\\c{r}}",
    0x0158: "{\\v{R}}",
    0x0159: "{\\v{r}}",
    0x015a: "{\\'S}",
    0x015b: "{\\'s}",
    0x015c: "{\\^S}",
    0x015d: "{\\^s}",
    0x015e: "{\\c{S}}",
    0x015f: "{\\c{s}}",
    0x0160: "{\\v{S}}",
    0x0161: "{\\v{s}}",
    0x0162: "{\\c{T}}",
    0x0163: "{\\c{t}}",
    0x0164: "{\\v{T}}",
    0x0165: "{\\v{t}}",
    0x0168: "{\\~U}",
    0x0169: "{\\~u}",
    0x016a: "{\\=U}",
    0x016b: "{\\=u}",
    0x016c: "{\\u{U}}",
    0x016d: "{\\u{u}}",
    0x016e: "{\\r{U}}",
    0x016f: "{\\r{u}}",
    0x0170: "{\\H{U}}",
    0x0171: "{\\H{u}}",
    0x0172: "{\\c{U}}",
    0x0173: "{\\c{u}}",
    0x0174: "{\\^W}",
    0x0175: "{\\^w}",
    0x0176: "{\\^Y}",
    0x0177: "{\\^y}",
    0x0178: '{\\"Y}',
    0x0179: "{\\'Z}",
    0x017a: "{\\'Z}",
    0x017b: "{\\.Z}",
    0x017c: "{\\.Z}",
    0x017d: "{\\v{Z}}",
    0x017e: "{\\v{z}}",

    0x01c4: "{D\\v{Z}}",
    0x01c5: "{D\\v{z}}",
    0x01c6: "{d\\v{z}}",
    0x01c7: "{LJ}",
    0x01c8: "{Lj}",
    0x01c9: "{lj}",
    0x01ca: "{NJ}",
    0x01cb: "{Nj}",
    0x01cc: "{nj}",
    0x01cd: "{\\v{A}}",
    0x01ce: "{\\v{a}}",
    0x01cf: "{\\v{I}}",
    0x01d0: "{\\v\\i}",
    0x01d1: "{\\v{O}}",
    0x01d2: "{\\v{o}}",
    0x01d3: "{\\v{U}}",
    0x01d4: "{\\v{u}}",
    0x01e6: "{\\v{G}}",
    0x01e7: "{\\v{g}}",
    0x01e8: "{\\v{K}}",
    0x01e9: "{\\v{k}}",
    0x01ea: "{\\c{O}}",
    0x01eb: "{\\c{o}}",
    0x01f0: "{\\v\\j}",
    0x01f1: "{DZ}",
    0x01f2: "{Dz}",
    0x01f3: "{dz}",
    0x01f4: "{\\'G}",
    0x01f5: "{\\'g}",
    0x01fc: "{\\'\\AE}",
    0x01fd: "{\\'\\ae}",
    0x01fe: "{\\'\\O}",
    0x01ff: "{\\'\\o}",

    0x02c6: '{\\^{}}',
    0x02dc: '{\\~{}}',
    0x02d8: '{\\u{}}',
    0x02d9: '{\\.{}}',
    0x02da: "{\\r{}}",
    0x02dd: '{\\H{}}',
    0x02db: '{\\c{}}',
    0x02c7: '{\\v{}}',

    0x03c0: '{\\mbox{$\\pi$}}',
    # consider adding more Greek here

    0xfb01: '{fi}',
    0xfb02: '{fl}',

    0x2013: '{--}',
    0x2014: '{---}',
    0x2018: "{`}",
    0x2019: "{'}",
    0x201c: "{``}",
    0x201d: "{''}",
    0x2020: "{\\dag}",
    0x2021: "{\\ddag}",
    0x2122: "{\\mbox{$^\\mbox{TM}$}}",
    0x2022: "{\\mbox{$\\bullet$}}",
    0x2026: "{\\ldots}",
    0x2202: "{\\mbox{$\\partial$}}",
    0x220f: "{\\mbox{$\\prod$}}",
    0x2211: "{\\mbox{$\\sum$}}",
    0x221a: "{\\mbox{$\\surd$}}",
    0x221e: "{\\mbox{$\\infty$}}",
    0x222b: "{\\mbox{$\\int$}}",
    0x2248: "{\\mbox{$\\approx$}}",
    0x2260: "{\\mbox{$\\neq$}}",
    0x2264: "{\\mbox{$\\leq$}}",
    0x2265: "{\\mbox{$\\geq$}}",
}
for _i in range(0x0020):
    if _i not in latex_equivalents:
        latex_equivalents[_i] = ''
for _i in range(0x0020, 0x007f):
    if _i not in latex_equivalents:
        latex_equivalents[_i] = chr(_i)

# Characters that should be ignored and not output in tokenization
_ignore = set([chr(i) for i in list(range(32)) + [127]]) - set('\t\n\r')

# Regexp of chars not in blacklist, for quick start of tokenize
_stoppers = re.compile('[\x00-\x1f!$\\-?\\{~\\\\`\']')

_blacklist = set(' \n\r')
_blacklist.add(None)    # shortcut candidate generation at end of data

# Construction of inverse translation table
_l2u = {
    r'\ ': ord(' ')  # unexpanding space makes no sense in non-TeX contexts
}

for _tex in latex_equivalents:
    if _tex <= 0x0020 or (_tex <= 0x007f and len(latex_equivalents[_tex]) <= 1):
        continue    # boring entry
    _toks = tuple(_tokenize(latex_equivalents[_tex]))
    if _toks[0] == '{' and _toks[-1] == '}':
        _toks = _toks[1:-1]
    if _toks[0].isalpha():
        continue    # don't turn ligatures into single chars
    if len(_toks) == 1 and (_toks[0] == "'" or _toks[0] == "`"):
        continue    # don't turn ascii quotes into curly quotes
    if _toks[0] == '\\mbox' and _toks[1] == '{' and _toks[-1] == '}':
        _toks = _toks[2:-1]
    if len(_toks) == 4 and _toks[1] == '{' and _toks[3] == '}':
        _toks = (_toks[0], _toks[2])
    if len(_toks) == 1:
        _toks = _toks[0]
    _l2u[_toks] = _tex

# Shortcut candidate generation for certain useless candidates:
# a character is in _blacklist if it can not be at the start
# of any translation in _l2u.  We use this to quickly skip through
# such characters before getting to more difficult-translate parts.
# _blacklist is defined several lines up from here because it must
# be defined in order to call _tokenize, however it is safe to
# delay filling it out until now.

for i in range(0x0020, 0x007f):
    _blacklist.add(chr(i))
_blacklist.remove('{')
_blacklist.remove('$')
for candidate in _l2u:
    if isinstance(candidate, tuple):
        if not candidate or not candidate[0]:
            continue  # pragma: no cover
        firstchar = candidate[0][0]
    else:
        firstchar = candidate[0]
    _blacklist.discard(firstchar)



"""
Common data model for clld applications.

We exploit several features of sqlalchemy to make the definition of the data model as
succinct as possible:
- custom declarative base class (see clld.db.meta)
- versioning (see clld.db.versioned)
- joined table inheritance (see clld.db.meta)
- mixins (see clld.db.models.common)

All mapper classes that must be available to attach app-specific custom behaviour are
marked as implementers of the corresponding interface.
"""



from clld.web.util.htmllib import HTML
from clld.web.util.helpers import external_link

__all__ = ['BASE_URL', 'url', 'logo', 'link']

BASE_URL = 'https://concepticon.clld.org'


def url(id=None, obj_type='ConceptSet'):
    if id is None:
        return BASE_URL
    path_map = {
        'ConceptSet': 'parameters',
        'ConceptList': 'contributions',
        'Concept': 'values',
    }
    return '{0}/{1}/{2}'.format(BASE_URL, path_map[obj_type], id)


def logo(req, width='20'):
    return HTML.img(src=req.static_url('clld:web/static/images/concepticon.png'), width=width)


def link(req, id=None, obj_type='ConceptSet', label=None):
    akw = dict(
        title='Concepticon' if id is None else 'View {0} at Concepticon'.format(obj_type),
        href=url(id=id, obj_type=obj_type))
    if label:
        return external_link(None, label=label, **akw)
    return HTML.a(logo(req), **akw)



import warnings
import importlib

import pytest
from zope.interface import Interface
from pyramid.testing import Configurator
from pyramid.httpexceptions import HTTPNotFound

from clld.db.models.common import Contribution, ValueSet, Language, Language_files
from clld.interfaces import IMapMarker, IDataTable
from clld.web.adapters.download import N3Dump


@pytest.fixture
def config():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            'ignore', category=DeprecationWarning, module='importlib._bootstrap')

        config_ = Configurator(
            root_package=importlib.import_module('clld.web'),
            settings={
                'sqlalchemy.url': 'sqlite://',
                'clld.pacific_centered_maps': True})
        config_.include('clld.web.app')
        return config_


def test_CLLDRequest(env):
    c = env['request'].db.query(Contribution).first()
    env['request'].resource_url(c, ext='geojson')
    assert env['request'].ctx_for_url('/some/path/to/nowhere') is None
    assert env['request'].ctx_for_url('/')
    env['request'].file_url(Language_files(id='1', object=Language.first()))
    assert env['request'].get_datatable('valuesets', ValueSet)
    assert env['request'].contact_email_address.startswith('from.settings')


def test_menu_item(env):
    from clld.web.app import menu_item

    assert menu_item('contributions', None, env['request'])


def test_ctx_factory(request_factory):
    from clld.web.app import ctx_factory

    for model, route in [
        (Contribution, 'contributions'),
        (ValueSet, 'valuesets'),
        (Language, 'languages'),
    ]:
        obj = model.first()

        with request_factory(matchdict={'id': obj.id}, matched_route=route) as req:
            ctx_factory(model, 'index', req)
            ctx_factory(model, 'rsc', req)

    with request_factory(matchdict={'id': 'xxx'}) as req:
        with pytest.raises(HTTPNotFound):
            ctx_factory(Contribution, 'rsc', req)


def test_MapMarker(env):
    marker = env['request'].registry.getUtility(IMapMarker)
    assert marker(None, env['request'])


def test_add_config_from_file(testsdir):
    from clld.web.app import add_settings_from_file

    config = Configurator()
    add_settings_from_file(config, testsdir / 'test.ini')
    assert 'app:main.use' in config.registry.settings


def test_config(config):
    class IF(Interface):
        pass

    # should have no effect, because a resource with this name is registered by
    # default:
    config.register_menu('languages', ('sources', dict(label='References')))
    config.register_resource('language', None, None)
    config.register_resource('testresource', Language, IF, with_index=True, test=True)
    config.register_download(N3Dump(Language, 'clld'))
    config.add_301('/301pattern', 'http://example.org')
    config.add_410('/410pattern')


def test_no_overwrite_registration(config):
    config.register_utility(1, IMapMarker)
    assert config.registry.queryUtility(IMapMarker) == 1
    config.register_utility(2, IMapMarker)
    assert config.registry.queryUtility(IMapMarker) == 2
    config.register_utility(3, IMapMarker, overwrite=False)
    assert config.registry.queryUtility(IMapMarker) == 2

    config.register_datatable('route', 1)
    assert config.registry.queryUtility(IDataTable, name='route') == 1
    config.register_datatable('route', 2, overwrite=False)
    assert config.registry.queryUtility(IDataTable, name='route') == 1


def test_includeme_error(tmp_path, capsys):
    import sys
    sys.path.append(str(tmp_path))
    pkg = tmp_path.joinpath('failingapp')
    pkg.mkdir()
    pkg.joinpath('__init__.py').write_text('#\n', 'ascii')
    pkg.joinpath('util.py').write_text('import xyzxyz', 'ascii')
    config = Configurator(
        root_package=importlib.import_module('failingapp'),
        settings={'sqlalchemy.url': 'sqlite://'})
    with pytest.raises(ImportError):
        config.include('clld.web.app')
    out, err = capsys.readouterr()
    assert 'failingapp.util' in out
    sys.path.pop()



from clld.web.util.concepticon import *


def test_url():
    assert url().endswith('concepticon.clld.org')
    assert url('1').endswith('parameters/1')
    assert url('x', obj_type='Concept').endswith('values/x')


def test_link(env):
    assert '.png' in link(env['request'], id='1')
    assert 'share' in link(env['request'], id='1', label='gloss')
